__all__ = ("MetricBundle", "create_empty_metric_bundle")

import os
import warnings
from copy import deepcopy

import numpy as np
import numpy.ma as ma

import rubin_sim.maf.maps as maps
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as stackers
import rubin_sim.maf.utils as utils
from rubin_sim.maf.stackers import ColInfo


def create_empty_metric_bundle():
    """Create an empty metric bundle.

    Returns
    -------
    MetricBundle : `MetricBundle`
        An empty metric bundle,
        configured with just the :class:`BaseMetric` and :class:`BaseSlicer`.
    """
    return MetricBundle(metrics.BaseMetric(), slicers.BaseSlicer(), "")


class MetricBundle:
    """Define a metric bundle combination of metric, slicer, and constraint.

    Parameters
    ----------
    metric : `~rubin_sim.maf.metric`
        The Metric class to run per slice_point
    slicer : `~rubin_sim.maf.slicer`
        The Slicer to apply to the incoming visit data (the observations).
    constraint : `str` or None, opt
        A (sql-style) constraint to apply to the visit data, to apply a
        broad sub-selection.
    stacker_list : `list` [`~rubin_sim.maf.stacker`], opt
        A list of pre-configured stackers to use to generate additional
        columns per visit.
        These will be generated automatically if needed, but pre-configured
        versions will override these.
    run_name : `str`, opt
        The name of the simulation being run.
        This will be added to output files and plots.
        Setting it prevents file conflicts when running the same
        metric on multiple simulations, and
        provides a way to identify which simulation is being analyzed.
    metadata : `str`, opt
        A deprecated version of info_label (below).
        Values set by metadata will be used for info_label.
        If both are set, info_label is used.
    info_label : `str` or None, opt
        Information to add to the output metric data file name and plot labels.
        If this is not provided, it will be auto-generated from the
        constraint (if any).
        Setting this provides an easy way to specify different
        configurations of a metric, a slicer,
        or just to rewrite your constraint into friendlier terms.
        (i.e. a constraint like 'note not like "%DD%"' can become
        "non-DD" in the file name and plot labels
        by specifying info_label).
    plot_dict : `dict` of plotting parameters, opt
        Specify general plotting parameters, such as x/y/color limits.
    display_dict : `dict` of display parameters, opt
        Specify parameters for show_maf web pages, such as the
        side bar labels and figure captions.
        Keys: 'group', 'subgroup', 'caption', and 'order'
        (such as to set metrics in filter order, etc)
    summary_metrics : `list` of `~rubin_sim.maf.metrics`
        A list of summary metrics to run to summarize the
        primary metric, such as MedianMetric, etc.
    maps_list : `list` of `~rubin_sim.maf.maps`
        A list of pre-configured maps to use for the metric.
        This will be auto-generated if specified
        by the metric class, but pre-configured versions will override these.

    Notes
    -----
    Define the "thing" you are measuring, with a combination of
    * metric (calculated per data_slice)
    * slicer (how to create the data_slices)
    * constraint (an optional definition of a large subset of data)

    Together these define a unique combination of an opsim benchmark,
    or "metric bundle".
    An example would be:
    a CountMetric, a HealpixSlicer, and a constraint of 'filter="r"'.

    After the metric is evaluated at each slice_point created by the
    slicer, the resulting metric values are saved in the MetricBundle.

    The MetricBundle also saves the summary metrics to be used
    to generate summary statistics over those metric values,
    as well as the resulting summary statistic values.

    Plotting parameters and display parameters (for show_maf) are saved
    in the MetricBundle, as well as additional info_label such as the
    opsim run name, and relevant stackers and maps
    to apply when calculating the metric values.
    """

    col_info = ColInfo()

    def __init__(
        self,
        metric,
        slicer,
        constraint=None,
        stacker_list=None,
        run_name="run name",
        metadata=None,
        info_label=None,
        plot_dict=None,
        display_dict=None,
        summary_metrics=None,
        maps_list=None,
        file_root=None,
        plot_funcs=None,
    ):
        # Set the metric.
        if not isinstance(metric, metrics.BaseMetric):
            raise ValueError("metric must be an rubin_sim.maf.metrics object")
        self.metric = metric
        # Set the slicer.
        if not isinstance(slicer, slicers.BaseSlicer):
            raise ValueError("slicer must be an rubin_sim.maf.slicers object")
        self.slicer = slicer
        # Set the constraint.
        self.constraint = constraint
        if self.constraint is None:
            self.constraint = ""
        # Set the stacker_list if applicable.
        if stacker_list is not None:
            if isinstance(stacker_list, stackers.BaseStacker):
                self.stacker_list = [
                    stacker_list,
                ]
            else:
                self.stacker_list = []
                for s in stacker_list:
                    if s is None:
                        pass
                    else:
                        if not isinstance(s, stackers.BaseStacker):
                            raise ValueError("stacker_list must only contain rubin_sim.maf.stackers objs")
                        self.stacker_list.append(s)
        else:
            self.stacker_list = []
        # Set the 'maps' to apply to the slicer, if applicable.
        if maps_list is not None:
            if isinstance(maps_list, maps.BaseMap):
                self.known_cols = [
                    maps_list,
                ]
            else:
                self.maps_list = []
                for m in maps_list:
                    if not isinstance(m, maps.BaseMap):
                        raise ValueError("maps_list must only contain rubin_sim.maf.maps objects")
                    self.maps_list.append(m)
        else:
            self.maps_list = []
        # If the metric knows it needs a particular map, add it to the list.
        map_names = [map_name.__class__.__name__ for map_name in self.maps_list]
        if hasattr(self.metric, "maps"):
            for map_needed in self.metric.maps:
                if isinstance(map_needed, str):
                    if map_needed not in map_names:
                        temp_map = getattr(maps, map_needed)()
                        self.maps_list.append(temp_map)
                        map_names.append(map_needed)
                else:
                    if map_needed not in self.maps_list:
                        self.maps_list.append(map_needed)
                        map_names.append(map_needed.__class__.__name__)

        # Add the summary stats, if applicable.
        self.set_summary_metrics(summary_metrics)
        # Set the provenance/info_label.
        self._build_metadata(info_label, metadata)
        # Set the run name and build the output filename base (fileRoot).
        self.set_run_name(run_name)
        # Reset fileRoot, if provided.
        if file_root is not None:
            self.file_root = file_root
        # Determine the columns needed from the database.
        self._find_req_cols()
        # Set the plotting classes/functions.
        self.set_plot_funcs(plot_funcs)
        # Set the plot_dict and displayDicts.
        self.plot_dict = {}
        self.set_plot_dict(plot_dict)
        # Update/set display_dict.
        self.display_dict = {}
        self.set_display_dict(display_dict)
        # This is where we store the metric values and summary stats.
        self.metric_values = None
        self.summary_values = None

    def _reset_metric_bundle(self):
        """Reset all properties of MetricBundle."""
        self.metric = None
        self.slicer = None
        self.constraint = None
        self.stacker_list = []
        self.summary_metrics = []
        self.plot_funcs = []
        self.maps_list = None
        self.run_name = "run name"
        self.info_label = ""
        self.db_cols = None
        self.file_root = None
        self.plot_dict = {}
        self.display_dict = {}
        self.metric_values = None
        self.summary_values = None

    def _setup_metric_values(self):
        """Set up the numpy masked array to store the metric value data."""
        dtype = self.metric.metric_dtype
        # Can't store healpix slicer mask values in an int array.
        if dtype == "int":
            dtype = "float"
        if self.metric.shape == 1:
            shape = self.slicer.shape
        else:
            shape = (self.slicer.shape, self.metric.shape)
        self.metric_values = ma.MaskedArray(
            data=np.empty(shape, dtype),
            mask=np.zeros(shape, "bool"),
            fill_value=self.slicer.badval,
        )
        if hasattr(self.slicer, "mask"):
            self.metric_values.mask = self.slicer.mask

    def _build_metadata(self, info_label, metadata=None):
        """If no info_label is provided, process the constraint
        (by removing extra spaces, quotes, the word 'filter' and equal signs)
        to make a info_label version.
        e.g. 'filter = "r"' becomes 'r'
        """
        # Pass the deprecated version into info_label if info_label is not set
        if metadata is not None and info_label is None:
            warnings.warn(
                'Use of "metadata" as a kwarg is deprecated, please use info_label instead.'
                f" (copying {metadata} into info_label). "
            )
            info_label = metadata
        if info_label is None:
            self.info_label = self.constraint.replace("=", "").replace("filter", "").replace("'", "")
            self.info_label = self.info_label.replace('"', "").replace("  ", " ")
            self.info_label = self.info_label.strip(" ")
        else:
            self.info_label = info_label

    def _build_file_root(self):
        """
        Build an auto-generated output filename root
        (i.e. minus the plot type or .npz ending).
        """
        # Build basic version.
        self.file_root = "_".join(
            [
                self.run_name,
                self.metric.name,
                self.info_label,
                self.slicer.slicer_name[:4].upper(),
            ]
        )
        # Sanitize output name if needed.
        self.file_root = utils.name_sanitize(self.file_root)

    def _find_req_cols(self):
        """Find the columns needed by the metrics, slicers, and stackers.
        If there are any additional stackers required, instatiate them
        and add them to the self.stackers list.
        (default stackers have to be instantiated to determine
        what additional columns are needed from database).
        """
        # Find all the columns needed by metric and slicer.
        known_cols = self.slicer.columns_needed + list(self.metric.col_name_arr)
        # For the stackers already set up, find their required columns.
        for s in self.stacker_list:
            known_cols += s.cols_req
        known_cols = set(known_cols)
        # Track sources of all of these columns.
        self.db_cols = set()
        new_stackers = set()
        for col in known_cols:
            if self.col_info.get_data_source(col) == self.col_info.default_data_source:
                self.db_cols.add(col)
            else:
                # New default stackers could come from metric/slicer
                # or stackers.
                new_stackers.add(self.col_info.get_data_source(col))
        # Remove already-specified stackers from default list.
        for s in self.stacker_list:
            if s.__class__ in new_stackers:
                new_stackers.remove(s.__class__)
        # Loop and check if stackers are introducing new columns or stackers.
        while len(new_stackers) > 0:
            # Check for the sources of the columns in any of the new stackers.
            new_cols = []
            for s in new_stackers:
                new_stacker = s()
                new_cols += new_stacker.cols_req
                self.stacker_list.append(new_stacker)
            new_cols = set(new_cols)
            new_stackers = set()
            for col in new_cols:
                if self.col_info.get_data_source(col) == self.col_info.default_data_source:
                    self.db_cols.add(col)
                else:
                    new_stackers.add(self.col_info.get_data_source(col))
            for s in self.stacker_list:
                if s.__class__ in new_stackers:
                    new_stackers.remove(s.__class__)
        # A Bit of cleanup.
        # Remove 'metricdata' from dbcols if it ended here by default.
        if "metricdata" in self.db_cols:
            self.db_cols.remove("metricdata")
        if "None" in self.db_cols:
            self.db_cols.remove("None")

    def set_summary_metrics(self, summary_metrics):
        """Set (or reset) the summary metrics for the metricbundle.

        Parameters
        ----------
        summary_metrics : `List` [`BaseMetric`]
            Instantiated summary metrics to use to calculate
            summary statistics for this metric.
        """
        if summary_metrics is not None:
            if isinstance(summary_metrics, metrics.BaseMetric):
                self.summary_metrics = [summary_metrics]
            else:
                self.summary_metrics = []
                for s in summary_metrics:
                    if not isinstance(s, metrics.BaseMetric):
                        raise ValueError("SummaryStats must only contain rubin_sim.maf.metrics objects")
                    self.summary_metrics.append(s)
        else:
            # Add identity metric to unislicer metric values
            # (to get them into resultsDB).
            if self.slicer.slicer_name == "UniSlicer":
                self.summary_metrics = [metrics.IdentityMetric()]
            else:
                self.summary_metrics = []

    def set_plot_funcs(self, plot_funcs):
        """Set or reset the plotting functions.

        The default is to use all the plotFuncs associated with the slicer,
        which is what happens in self.plot if setPlotFuncs is not used to
        override self.plotFuncs.

        Parameters
        ----------
        plot_funcs : `List` [`BasePlotter`]
            The plotter or plotters to use to generate visuals for this metric.
        """
        if plot_funcs is not None:
            if plot_funcs is isinstance(plot_funcs, plots.BasePlotter):
                self.plot_funcs = [plot_funcs]
            else:
                self.plot_funcs = []
                for p_func in plot_funcs:
                    if not isinstance(p_func, plots.BasePlotter):
                        raise ValueError(
                            "plotFuncs should contain instantiated " + "rubin_sim.maf.plotter objects."
                        )
                    self.plot_funcs.append(p_func)
        else:
            self.plot_funcs = []
            for p_func in self.slicer.plot_funcs:
                if isinstance(p_func, plots.BasePlotter):
                    self.plot_funcs.append(p_func)
                else:
                    self.plot_funcs.append(p_func())

    def set_plot_dict(self, plot_dict):
        """Set or update any property of plot_dict.

        Parameters
        ----------
        plot_dict : dict
            A dictionary of plotting parameters.
            The usable keywords vary with each rubin_sim.maf.plots Plotter.
        """
        # Don't auto-generate anything here - the plotHandler does it.
        if plot_dict is not None:
            self.plot_dict.update(plot_dict)
        # Check for bad zp or norm_val values.
        if "zp" in self.plot_dict:
            if self.plot_dict["zp"] is not None:
                if not np.isfinite(self.plot_dict["zp"]):
                    warnings.warn(
                        "Warning! Plot zp for %s was infinite: removing zp from plot_dict" % (self.file_root)
                    )
                    del self.plot_dict["zp"]
        if "norm_val" in self.plot_dict:
            if self.plot_dict["norm_val"] == 0:
                warnings.warn(
                    "Warning! Plot normalization value for %s was 0: removing norm_val from plot_dict"
                    % (self.file_root)
                )
                del self.plot_dict["norm_val"]

    def set_display_dict(self, display_dict=None, results_db=None):
        """Set or update any property of display_dict.

        Parameters
        ----------
        display_dict : `dict` of `str`
            Dictionary of display parameters for show_maf.
            Expected keys are: 'group', 'subgroup', 'order', 'caption'.
            'group', 'subgroup', and 'order' control where the
            metric results are shown on the show_maf page.
            'caption' provides a caption to use with the metric results.
            These values are saved in the results database.
        results_db : `maf.ResultsDb`
            A MAF results database, used to save the display parameters.
        """
        # Set up a temporary dictionary with the default values.
        tmp_display_dict = {
            "group": None,
            "subgroup": None,
            "order": 0,
            "caption": None,
        }
        # Update from self.display_dict (to use existing values, if present).
        tmp_display_dict.update(self.display_dict)
        # And then update from any values being passed now.
        if display_dict is not None:
            tmp_display_dict.update(display_dict)
        # Reset self.display_dict to this updated dictionary.
        self.display_dict = tmp_display_dict
        # If we still need to auto-generate a caption, do it.
        if self.display_dict["caption"] is None:
            if self.metric.comment is None:
                caption = self.metric.name + " calculated on a %s basis" % (self.slicer.slicer_name)
                if self.constraint != "" and self.constraint is not None:
                    caption += " using a subset of data selected via %s." % (self.constraint)
                else:
                    caption += "."
            else:
                caption = self.metric.comment
            if "zp" in self.plot_dict:
                caption += " Values plotted with a zeropoint of %.2f." % (self.plot_dict["zp"])
            if "norm_val" in self.plot_dict:
                caption += " Values plotted with a normalization value of %.2f." % (
                    self.plot_dict["norm_val"]
                )
            self.display_dict["caption"] = caption
        if results_db:
            # Update the display values in the results_db.
            metric_id = results_db.update_metric(
                self.metric.name,
                self.slicer.slicer_name,
                self.run_name,
                self.constraint,
                self.info_label,
                None,
            )
            results_db.update_display(metric_id, self.display_dict)

    def set_run_name(self, run_name, update_file_root=True):
        """Set (or reset) the run_name.
        FileRoot will be updated accordingly if desired.

        Parameters
        ----------
        run_name : `str`
            Run Name, which will become part of the fileRoot.
        fileRoot : `bool`, optional
            Flag to update the fileRoot with the run_name.
        """
        self.run_name = run_name
        if update_file_root:
            self._build_file_root()

    def write_db(self, results_db=None, outfile_suffix=None):
        """Write the metric information to the results database"""
        if outfile_suffix is not None:
            outfile = self.file_root + "_" + outfile_suffix + ".npz"
        else:
            outfile = self.file_root + ".npz"
        if results_db is not None:
            metric_id = results_db.update_metric(
                self.metric.name,
                self.slicer.slicer_name,
                self.run_name,
                self.constraint,
                self.info_label,
                outfile,
            )
            results_db.update_display(metric_id, self.display_dict)

    def write(self, comment="", out_dir=".", outfile_suffix=None, results_db=None):
        """Write metric_values (and associated info_label) to disk.

        Parameters
        ----------
        comment : `str`
            Any additional comments to add to the output file
        out_dir : `str`
            The output directory
        outfile_suffix :`str`
            Additional suffix to add to the output files
            (typically a numerical suffix for movies)
        results_db : `maf.ResultsDb`
            Results database to store information on the file output
        """
        if outfile_suffix is not None:
            outfile = self.file_root + "_" + outfile_suffix + ".npz"
        else:
            outfile = self.file_root + ".npz"
        self.slicer.write_data(
            os.path.join(out_dir, outfile),
            self.metric_values,
            metric_name=self.metric.name,
            sim_data_name=self.run_name,
            constraint=self.constraint,
            info_label=self.info_label + comment,
            display_dict=self.display_dict,
            plot_dict=self.plot_dict,
            summary_values=self.summary_values,
        )
        if results_db is not None:
            self.write_db(results_db=results_db)

    def output_json(self):
        """Set up and call the baseSlicer outputJSON method,
        to output to IO string.

        Returns
        -------
        io
           IO object containing JSON data representing the metric bundle data.
        """
        io = self.slicer.output_json(
            self.metric_values,
            metric_name=self.metric.name,
            simDataName=self.run_name,
            info_label=self.info_label,
            plot_dict=self.plot_dict,
        )
        return io

    def read(self, filename):
        """Read metric data from disk.
        Overwrites any data currently in metricbundle.

        Parameters
        ----------
        filename : `str`
           The file from which to read the metric bundle data.
        """
        if not os.path.isfile(filename):
            raise IOError("%s not found" % filename)

        self._reset_metric_bundle()
        # Set up a base slicer to read data (we don't know type yet).
        baseslicer = slicers.BaseSlicer()
        # Use baseslicer to read file.
        metric_values, slicer, header = baseslicer.read_data(filename)
        self.slicer = slicer
        self.metric_values = metric_values
        self.metric_values.fill_value = slicer.badval
        # It's difficult to reinstantiate the metric object, as we don't
        # know what it is necessarily -- the metric_name can be changed.
        self.metric = metrics.BaseMetric()
        # But, for plot label building, we do need to try to recreate the
        #  metric name and units.
        self.metric.units = ""
        if header is not None:
            self.metric.name = header["metric_name"]
            if "plot_dict" in header:
                if "units" in header["plot_dict"]:
                    self.metric.units = header["plot_dict"]["units"]
            self.run_name = header["sim_data_name"]
            try:
                self.constraint = header["constraint"]
            except KeyError:
                self.constraint = header["sqlconstraint"]
            # and then use info_label if it's there instead
            if "info_label" in header:
                self.info_label = header["info_label"]
            if "plot_dict" in header:
                self.set_plot_dict(header["plot_dict"])
            if "display_dict" in header:
                self.set_display_dict(header["display_dict"])
            if "summary_values" in header:
                self.summary_values = header["summary_values"]
        if self.info_label is None:
            self._build_metadata()
        path, head = os.path.split(filename)
        self.file_root = head.replace(".npz", "")
        self.set_plot_funcs(None)

    @classmethod
    def load(cls, filename):
        """Create a metric bundle and load its content from disk.

        Parameters
        ----------
        filename : `str`
           The file from which to read the metric bundle data.
        """
        metric_bundle = cls(metrics.BaseMetric(), slicers.BaseSlicer(), "")
        metric_bundle.read(filename)
        return metric_bundle

    def compute_summary_stats(self, results_db=None):
        """Compute summary statistics on metric_values,
        using summaryMetrics (metricbundle list).

        Parameters
        ----------
        results_db : Optional[ResultsDb]
            ResultsDb object to use to store the
            summary statistic values on disk.
        """
        if self.summary_values is None:
            self.summary_values = {}
        if self.summary_metrics is not None:
            # Build array of metric values, to use for  summary statistics.
            rarr_std = np.array(
                list(zip(self.metric_values.compressed())),
                dtype=[("metricdata", self.metric_values.dtype)],
            )
            for m in self.summary_metrics:
                # The summary metric colname should already be set
                # to 'metricdata', but in case it's not:
                m.colname = "metricdata"
                summary_name = m.name.replace(" metricdata", "").replace(" None", "")
                if hasattr(m, "mask_val"):
                    # summary metric requests to use the mask value,
                    # as specified by itself,
                    # rather than skipping masked vals.
                    rarr = np.array(
                        list(zip(self.metric_values.filled(m.mask_val))),
                        dtype=[("metricdata", self.metric_values.dtype)],
                    )
                else:
                    rarr = rarr_std
                if np.size(rarr) == 0:
                    summary_val = self.slicer.badval
                else:
                    summary_val = m.run(rarr)
                self.summary_values[summary_name] = summary_val
                # Add summary metric info to results database, if applicable.
                if results_db:
                    metric_id = results_db.update_metric(
                        self.metric.name,
                        self.slicer.slicer_name,
                        self.run_name,
                        self.constraint,
                        self.info_label,
                        None,
                    )
                    results_db.update_summary_stat(
                        metric_id, summary_name=summary_name, summary_value=summary_val
                    )

    def reduce_metric(
        self,
        reduce_func,
        reduce_func_name=None,
        reduce_plot_dict=None,
        reduce_display_dict=None,
    ):
        """Run 'reduceFunc' (any function that operates on self.metric_values).

        Typically reduceFunc will be the metric reduce functions,
        as they are tailored to expect the metric_values format.
        reduceDisplayDict and reducePlotDicts are displayDicts
        and plotDicts to be applied to the new metricBundle.

        Parameters
        ----------
        reduce_func : `Func`
            Any function that will operate on self.metric_values
            (typically metric.reduce* function).
        reduce_plot_dict : `dict`, opt
            Plot dictionary for the results of the reduce function.
        reduce_display_dict : `dict`, opt
            Display dictionary for the results of the reduce function.

        Returns
        -------
        newmetric_bundle: `MetricBundle`
           New metric bundle, inheriting info_label from this metric bundle,
           but containing the new metric values calculated with
           the 'reduceFunc'.
        """
        # Generate a name for the metric values processed by the reduceFunc.
        if reduce_func_name is not None:
            r_name = reduce_func_name.replace("reduce_", "")
        else:
            r_name = reduce_func.__name__.replace("reduce_", "")
        reduce_name = self.metric.name + "_" + r_name
        # Set up metricBundle to store new metric values,
        # and add plot_dict/display_dict.
        newmetric = deepcopy(self.metric)
        newmetric.name = reduce_name
        newmetric.metric_dtype = "float"
        if reduce_plot_dict is not None:
            if "units" in reduce_plot_dict:
                newmetric.units = reduce_plot_dict["units"]
        newmetric_bundle = MetricBundle(
            metric=newmetric,
            slicer=self.slicer,
            stacker_list=self.stacker_list,
            constraint=self.constraint,
            info_label=self.info_label,
            run_name=self.run_name,
            plot_dict=None,
            plot_funcs=self.plot_funcs,
            display_dict=None,
            summary_metrics=self.summary_metrics,
            maps_list=self.maps_list,
            file_root="",
        )
        # Build a new output file root name.
        newmetric_bundle._build_file_root()
        # Add existing plot_dict (except for title/xlabels etc)
        # into new plot_dict.
        for k, v in self.plot_dict.items():
            if k not in newmetric_bundle.plot_dict:
                newmetric_bundle.plot_dict[k] = v
        # Update newmetric_bundle's plot dictionary with
        # any set explicitly by reducePlotDict.
        newmetric_bundle.set_plot_dict(reduce_plot_dict)
        # Copy the parent metric's display dict into the reduce display dict.
        newmetric_bundle.set_display_dict(self.display_dict)
        # Set the reduce function display 'order'
        # (this is set in the BaseMetric by default,
        # but can be overriden in a metric).
        order = newmetric.reduce_order[r_name]
        newmetric_bundle.display_dict["order"] = order
        # And then update the newmetric_bundle's
        # display dictionary with any set
        # explicitly by reduceDisplayDict.
        newmetric_bundle.set_display_dict(reduce_display_dict)
        # Set up new metricBundle's metric_values masked arrays,
        # copying metricValue's mask.
        newmetric_bundle.metric_values = ma.MaskedArray(
            data=np.empty(len(self.slicer), "float"),
            mask=self.metric_values.mask.copy(),
            fill_value=self.slicer.badval,
        )
        # Fill the reduced metric data using the reduce function.
        for i, (mVal, mMask) in enumerate(zip(self.metric_values.data, self.metric_values.mask)):
            if not mMask:
                val = reduce_func(mVal)
                newmetric_bundle.metric_values.data[i] = val
                if val == newmetric.badval:
                    newmetric_bundle.metric_values.mask[i] = True

        return newmetric_bundle

    def plot(self, plot_handler=None, plot_func=None, outfile_suffix=None, savefig=False):
        """
        Create all plots available from the slicer.
        plotHandler holds the output directory info, etc.

        Parameters
        ----------
        plot_handler : `~maf.plots.plot_handler`, opt
           The plot_handler saves the output location
           and results_db connection for a set of plots.
        plot_func : `maf.plots.BasePlotter`, opt
           Any plotter function.
           If not specified, the plotters in self.plotFuncs will be used.
        outfile_suffix : `str`, opt
           Optional string to append to the end of the plot output files.
           Useful when creating sequences of images for movies.
        savefig : `bool`, opt
           Flag indicating whether or not to save the figure to disk.
           Default is False.

        Returns
        -------
        made_plots : `dict`
            Dictionary of plot_type:figure key/value pairs,
            indicating what plots were created
            and what matplotlib figures were used.
        """
        # Generate a plot_handler if none was set.
        if plot_handler is None:
            plot_handler = plots.PlotHandler(savefig=savefig)
        # Make plots.
        if plot_func is not None:
            if isinstance(plot_func, plots.BasePlotter):
                plot_func = plot_func
            else:
                plot_func = plot_func()

        plot_handler.set_metric_bundles([self])
        plot_handler.set_plot_dicts(plot_dicts=[self.plot_dict], reset=True)
        made_plots = {}
        if plot_func is not None:
            fig = plot_handler.plot(plot_func, outfile_suffix=outfile_suffix)
            made_plots[plot_func.plotType] = fig
        else:
            for plot_func in self.plot_funcs:
                fig = plot_handler.plot(plot_func, outfile_suffix=outfile_suffix)
                made_plots[plot_func.plot_type] = fig
        return made_plots



__all__ = ("run_selfcal_metric",)

import argparse
import os
import shutil

import healpy as hp
import matplotlib
import numpy as np

matplotlib.use("Agg")

from .db import ResultsDb
from .maf_contrib import PhotometricSelfCalUniformityMetric
from .metric_bundles import MetricBundle, MetricBundleGroup
from .metrics import IdentityMetric
from .plots import HealpixHistogram, HealpixSkyMap, PlotHandler
from .slicers import HealpixSlicer, UniSlicer


def run_selfcal_metric():
    """
    Run the self-calibration metric on one database.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--db", type=str, default=None)
    parser.add_argument(
        "--no_clobber",
        dest="no_clobber",
        default=False,
        action="store_true",
        help="Do not remove existing directory outputs",
    )
    parser.add_argument("--filter", type=str, default="r")

    args = parser.parse_args()

    opsdb = args.db
    sim_name = os.path.basename(opsdb).replace(".db", "")

    out_dir = sim_name + "_selfcal"
    if not args.no_clobber:
        if os.path.isdir(out_dir):
            shutil.rmtree(out_dir)

    # Set up the metric bundle.
    map_nside = 128
    selfcal_metric = PhotometricSelfCalUniformityMetric(
        nside_residual=map_nside, filter_name=args.filter, metric_name="selfcal_uniformity-%s" % args.filter
    )
    slicer = UniSlicer()
    # Exclude DDF visits
    sql = "scheduler_note not like '%DD%'"
    # And run on only year 1 (?)
    sql += " and night < 366"
    sql += " and filter='%s'" % args.filter

    bundle = MetricBundle(
        selfcal_metric, slicer, sql, run_name=sim_name, info_label="year 1 no-DD %s" % args.filter
    )

    # Set up the resultsDB
    results_db = ResultsDb(out_dir=out_dir)
    # Go and run it
    group = MetricBundleGroup(
        {"selfcal": bundle}, opsdb, out_dir=out_dir, results_db=results_db, save_early=True
    )
    group.run_all(clear_memory=False, plot_now=True)

    # Make plots of the residuals map
    map_bundle = MetricBundle(
        IdentityMetric(metric_name="PhotoCal Uniformity %s" % args.filter),
        HealpixSlicer(map_nside),
        sql,
        run_name=sim_name,
        info_label="year 1 no-DD",
    )
    tmp_vals = bundle.metric_values[0]["uniformity_map"]
    tmp_vals = np.where(tmp_vals == hp.UNSEEN, map_bundle.slicer.badval, tmp_vals)
    map_bundle.metric_values = np.ma.MaskedArray(
        data=tmp_vals,
        mask=np.zeros(map_bundle.slicer.shape, "bool"),
        fill_value=map_bundle.slicer.badval,
    )
    map_bundle.write(out_dir=out_dir, results_db=results_db)

    ph = PlotHandler(results_db=results_db, out_dir=out_dir)
    ph.set_metric_bundles([map_bundle])
    _ = ph.plot(HealpixSkyMap(), plot_dicts={"color_min": -0.02, "color_max": 0.02})
    _ = ph.plot(HealpixHistogram(), plot_dicts={"percentile_clip": 99})

    results_db.close()



__all__ = ("make_fbs_tracking_db",)

import argparse
import os
import sqlite3

import pandas as pd

from rubin_sim.maf.db import ResultsDb, TrackingDb, VersionRow


def make_fbs_tracking_db():
    """Make a database to track lots of opsim outputs,
    adding in comments and identifiers based on the metric subsets.
    """
    parser = argparse.ArgumentParser(
        description="Create a tracking database for many opsim maf outputs,"
        "with comments and dates. "
        "Assumes metrics are in a subdirectory called 'maf'."
    )
    parser.add_argument(
        "--opsim_list",
        type=str,
        default="opsim_list",
        help="File containing a list of all of the opsim runs to add.",
    )
    args = parser.parse_args()

    batches = {
        "meta": "General Info Metrics",
        "glance": "Quick Look Metrics",
        "sci": "Science Metrics",
        "ss": "Solar System Metrics",
        "ddf": "Deep Drilling Metrics",
    }

    tracking_db = TrackingDb()
    print(f"Tracking database in directory {tracking_db.tracking_db_dir}")

    with open(args.opsim_list, "r") as runlist:
        for run in runlist:
            db_file = run.replace("\n", "")
            vals = db_file.split("/")
            family = vals[-2]
            db_name = vals[-1]
            run_name = db_name.replace(".db", "")
            run_version = run_name.split("_10yrs")[0][-4:]
            run_group = run_version + family

            # Try to build a comment on the run based on the run_name
            run_comment = run_name.replace("_10yrs", "")[0:-4]
            run_comment = run_comment.replace("_", "")

            print(run_name, db_file)
            conn = sqlite3.connect(db_file)
            query = "select Value from info where Parameter == 'Date, ymd'"
            result = pd.read_sql(query, conn)
            sched_date = result.iloc[0, 0]
            query = "select Value from info where Parameter like 'rubin_%.__version__'"
            result = pd.read_sql(query, conn)
            sched_version = result.iloc[0, 0]

            # Look for metrics in any of the above sets
            for k in batches:
                maf_dir = os.path.join("maf", run_name + "_" + k)
                maf_comment = batches[k]
                # Get maf run date and version
                if os.path.isfile(os.path.join(maf_dir, "resultsDb_sqlite.db")):
                    resdb = ResultsDb(maf_dir)
                    resdb.open()
                    query = resdb.session.query(VersionRow).all()
                    for v in query:
                        maf_version = v.version
                        maf_date = v.run_date
                    resdb.close()

                    maf_dir = os.path.relpath(maf_dir, start=os.path.dirname(tracking_db.tracking_db_dir))
                    runId = tracking_db.add_run(
                        run_group=run_group,
                        run_name=run_name,
                        run_comment=run_comment,
                        run_version=sched_version,
                        run_date=sched_date,
                        maf_comment=maf_comment,
                        maf_version=maf_version,
                        maf_date=maf_date,
                        maf_dir=maf_dir,
                        db_file=db_file,
                    )
                    print("Used MAF RunID %d" % (runId))



__all__ = ("FOPlot", "SummaryHistogram")


import healpy as hp
import matplotlib.pyplot as plt
import numpy as np

import rubin_sim.maf.metrics as metrics

from .plot_handler import BasePlotter


class FOPlot(BasePlotter):
    """
    Special plotter to generate and label fO plots.
    """

    def __init__(self):
        self.plot_type = "FO"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": "Number of visits",
            "ylabel": "Area (1000s of square degrees)",
            "scale": None,
            "asky": 18000.0,
            "n_visits": 825,
            "x_min": 0,
            "x_max": None,
            "y_min": 0,
            "y_max": None,
            "linewidth": 2,
            "reflinewidth": 2,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """
        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure number to use. Default = None, starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           Figure with the plot.
        """
        if not hasattr(slicer, "nside"):
            raise ValueError("FOPlot to be used with healpix or healpix derived slicers.")

        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)

        if plot_dict["scale"] is None:
            plot_dict["scale"] = hp.nside2pixarea(slicer.nside, degrees=True) / 1000.0

        if fig is None:
            fig, ax = plt.subplots(figsize=plot_dict["figsize"])

        # Expect metric_value to be something like number of visits
        cumulative_area = np.arange(1, metric_value.compressed().size + 1)[::-1] * plot_dict["scale"]
        ax.plot(
            np.sort(metric_value.compressed()),
            cumulative_area,
            "k-",
            linewidth=plot_dict["linewidth"],
            zorder=0,
        )
        # This results in calculating the summary stats in two places ..
        # not the ideal choice but easiest for most uses in this case.
        asky = plot_dict["asky"]
        n_visits = plot_dict["n_visits"]
        rarr = np.array(list(zip(metric_value.compressed())), dtype=[("fO", metric_value.dtype)])
        f_o_area = metrics.FOArea(col="fO", n_visit=n_visits, norm=False, nside=slicer.nside).run(rarr)
        f_o_nv = metrics.FONv(col="fO", asky=asky, norm=False, nside=slicer.nside).run(rarr)

        ax.axvline(x=n_visits, linewidth=plot_dict["reflinewidth"], color="b", linestyle=":")
        ax.axhline(y=asky / 1000.0, linewidth=plot_dict["reflinewidth"], color="r", linestyle=":")
        # Add lines for nvis_median and f_o_area:
        # note if these are -666 (badval), they will 'disappear'
        nvis_median = f_o_nv["value"][np.where(f_o_nv["name"] == "MedianNvis")][0]

        ax.axvline(
            x=nvis_median,
            linewidth=plot_dict["reflinewidth"],
            color="b",
            alpha=0.5,
            linestyle="-",
            label=f"f$_0$ Med. Nvis. (@ {asky/1000 :.0f}K sq deg) = {nvis_median :.0f} visits",
        )
        ax.axhline(
            y=f_o_area / 1000.0,
            linewidth=plot_dict["reflinewidth"],
            color="r",
            alpha=0.5,
            linestyle="-",
            label=f"f$_0$ Area (@ {n_visits :.0f} visits) = {f_o_area/1000 :.01f}K sq deg",
        )
        ax.legend(loc="upper right", fontsize="small", numpoints=1, framealpha=1.0)

        ax.set_xlabel(plot_dict["xlabel"])
        ax.set_ylabel(plot_dict["ylabel"])
        ax.set_title(plot_dict["title"])

        x_min = plot_dict["x_min"]
        x_max = plot_dict["x_max"]
        y_min = plot_dict["y_min"]
        y_max = plot_dict["y_max"]
        if (x_min is not None) or (x_max is not None):
            ax.set_xlim([x_min, x_max])
        if (y_min is not None) or (y_max is not None):
            ax.set_ylim([y_min, y_max])
        return fig


class SummaryHistogram(BasePlotter):
    """
    Special plotter to summarize metrics which return a set of values
    at each slice_point, e.g. a histogram the metric result per slicepoint.
    (example: the results of with the rubin_sim.maf.metrics.TgapsMetric).
    Essentially, this collapses the metric value over the sky and
    plots a summarized version (reduced to a single value per point
    according to the plot_dict['metricReduce'] metric).
    """

    def __init__(self):
        self.plot_type = "SummaryHistogram"
        self.object_plotter = True
        self.default_plot_dict = {
            "title": None,
            "xlabel": None,
            "ylabel": "Count",
            "label": None,
            "cumulative": False,
            "x_min": None,
            "x_max": None,
            "y_min": None,
            "y_max": None,
            "color": "b",
            "linestyle": "-",
            "histStyle": True,
            "grid": True,
            "metricReduce": metrics.SumMetric(),
            "yscale": None,
            "xscale": None,
            "bins": None,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """
        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            Handles 'object' datatypes for the masked array.
        slicer : `rubin_sim.maf.slicer`
            Any MAF slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user to override defaults.
            'metricReduce' (a `rubin_sim.maf.metric`) indicates how to
            marginalize the metric values calculated at each point to a
            single series of values over the sky.
            'histStyle' (True/False) indicates whether to plot the
            results as a step histogram (True) or as a series of values (False)
            'bins' (np.ndarray) sets the x values for the resulting plot
            and should generally match the bins used with the metric.
        fig : `matplotlib.figure.Figure`
            Matplotlib figure to use. Default starts a new figure.

        Returns
        -------
        fig: `matplotlib.figure.Figure`
           Matplotlib figure used to create the plot.
        """
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        # Combine the metric values across all slice_points.
        if not isinstance(plot_dict["metricReduce"], metrics.BaseMetric):
            raise ValueError("Expected plot_dict[metricReduce] to be a MAF metric object.")
        # Check that there is data to plot
        if np.size(metric_value.compressed()) == 0:
            raise ValueError(f"Did not find any data to plot in {self.plot_type}.")
        # Get the data type
        dt = metric_value.compressed()[0].dtype
        # Change an array of arrays to a 2-d array of correct dtype
        m_v = np.array(metric_value.compressed().tolist(), dtype=[("metric_value", dt)])
        # Make an array to hold the combined result
        final_hist = np.zeros(m_v.shape[1], dtype=float)
        metric = plot_dict["metricReduce"]
        metric.colname = "metric_value"
        # Loop over each bin and use the selected metric to combine the results
        for i in np.arange(final_hist.size):
            final_hist[i] = metric.run(m_v[:, i])

        if plot_dict["cumulative"]:
            final_hist = final_hist.cumsum()

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        bins = plot_dict["bins"]
        if plot_dict["histStyle"]:
            leftedge = bins[:-1]
            rightedge = bins[1:]

            x = np.vstack([leftedge, rightedge]).T.flatten()
            y = np.vstack([final_hist, final_hist]).T.flatten()

        else:
            # Could use this to plot things like FFT
            x = bins[:-1]
            y = final_hist
        # Make the plot.
        plt.plot(
            x,
            y,
            linestyle=plot_dict["linestyle"],
            label=plot_dict["label"],
            color=plot_dict["color"],
        )
        # Add labels.
        plt.xlabel(plot_dict["xlabel"])
        plt.ylabel(plot_dict["ylabel"])
        plt.title(plot_dict["title"])
        plt.grid(plot_dict["grid"], alpha=0.3)
        # Set y and x limits, if provided.
        if plot_dict["x_min"] is not None:
            plt.xlim(left=plot_dict["x_min"])
        elif bins[0] == 0:
            plt.xlim(left=0)
        if plot_dict["x_max"] is not None:
            plt.xlim(right=plot_dict["x_max"])
        if plot_dict["y_min"] is not None:
            plt.ylim(bottom=plot_dict["y_min"])
        elif final_hist.min() == 0:
            plot_dict["y_min"] = 0
        if plot_dict["y_max"] is not None:
            plt.ylim(top=plot_dict["y_max"])

        if plot_dict["yscale"] is not None:
            plt.yscale(plot_dict["yscale"])
        if plot_dict["xscale"] is not None:
            plt.xscale(plot_dict["xscale"])

        return fig



__all__ = (
    "LcfastNew",
    "LoadReference",
    "GetReference",
    "SnRate",
    "CovColor",
    "load_sne_cached",
)

import os
import warnings

import numpy as np
import pandas as pd
from astropy.cosmology import FlatLambdaCDM
from astropy.table import Table
from rubin_scheduler.data import get_data_dir
from scipy.interpolate import RegularGridInterpolator

from rubin_sim.maf.utils import m52snr

warnings.filterwarnings("ignore", category=RuntimeWarning)

steradian2_sqdeg = 180.0**2 / np.pi**2
# Mpc^3 -> Mpc^3/sr
norm = 1.0 / (4.0 * np.pi)


class LcfastNew:
    """
    class to simulate supernovae light curves in a fast way
    The method relies on templates and broadcasting to increase speed

    Parameters
    ---------------
    reference_lc : `scipy.interpolate.RegularGridData`
        lc reference files
    x1 : `float`
        SN stretch
    color : `float`
        SN color
    telescope : Telescope()
        telescope for the study
    mjd_col : `str`, optional
        name of the MJD col in data to simulate
    filter_col : `str`, optional
        name of the filter col in data to simulate
    exptime_col : `str`, optional
        name of the exposure time  col in data to simulate
    m5_col : `str`, optional
        name of the fiveSigmaDepth col in data to simulate
    season_col : `str`, optional
        name of the season col in data to simulate
    snr_min : `float`, optional
        minimal Signal-to-Noise Ratio to apply on LC points
    light_output : `bool`, optional
        to get a lighter output (ie lower number of cols)
    bluecutoff : `float`, optional
        blue cutoff for SN
    redcutoff : `float`, optional
        red cutoff for SN
    """

    def __init__(
        self,
        reference_lc,
        x1,
        color,
        mjd_col="observationStartMJD",
        filter_col="filter",
        exptime_col="visitExposureTime",
        m5_col="fiveSigmaDepth",
        season_col="season",
        nexp_col="numExposures",
        seeing_col="seeingFwhmEff",
        snr_min=5.0,
        light_output=True,
        ebvof_mw=-1.0,
        bluecutoff=380.0,
        redcutoff=800.0,
    ):
        # grab all vals
        self.filter_col = filter_col
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.exptime_col = exptime_col
        self.season_col = season_col
        self.nexp_col = nexp_col
        self.seeing_col = seeing_col

        self.x1 = x1
        self.color = color
        self.light_output = light_output
        self.ebvof_mw = ebvof_mw

        # Loading reference file
        self.reference_lc = reference_lc

        # This cutoffs are used to select observations:
        # phase = (mjd - DayMax)/(1.+z)
        # selection: min_rf_phase < phase < max_rf_phase
        # and        blue_cutoff < mean_rest_frame < red_cutoff
        # where mean_rest_frame = telescope.mean_wavelength/(1.+z)
        self.blue_cutoff = bluecutoff
        self.red_cutoff = redcutoff

        # SN parameters for Fisher matrix estimation
        self.param__fisher = ["x0", "x1", "daymax", "color"]

        self.snr_min = snr_min

    def __call__(self, obs, ebvof_mw, gen_par=None, bands="grizy"):
        """Simulation of the light curve


        Parameters
        -----------
        obs : `np.ndarray`, (N,)
            array of observations
        ebvof_mw : `float`
            E(B-V) for MW
        gen_par : `np.ndarray`, optional
            simulation parameters (default: None)
        bands : `str`, optional
            filters to consider for simulation (default: grizy)


        Returns
        --------
        table : `astropy.Table`
            columns: band, flux, fluxerr, snr_m5,flux_e,zp,zpsys,time
            metadata : SNID,RA,Dec,DayMax,X1,Color,z
        """

        if len(obs) == 0:
            return None

        # result in this df
        tab_tot = pd.DataFrame()

        # loop on the bands
        for band in bands:
            idx = obs[self.filter_col] == band
            if len(obs[idx]) > 0:
                tab_tot = pd.concat(
                    [tab_tot, self.process_band(obs[idx], ebvof_mw, band, gen_par)],
                    ignore_index=True,
                )

        # return produced LC
        return tab_tot

    def process_band(self, sel_obs, ebvof_mw, band, gen_par):
        """LC simulation of a set of obs corresponding to a band
        The idea is to use python broadcasting so as to estimate
        all the requested values (flux, flux error, Fisher components, ...)
        in a single path (i.e no loop!)

        Parameters
        ---------------
        sel_obs : `np.ndarray`
            array of observations
        band : `str`
            band of observations
        gen_par : `np.ndarray``
            simulation parameters

        Returns
        -------
        table : `astropy.Table`
            astropy table with fields corresponding to LC components
        """

        # if there are no observations in this filter: return None
        if len(sel_obs) == 0:
            return None

        # Get the fluxes (from griddata reference)

        # xi = MJD-T0
        xi = sel_obs[self.mjd_col] - gen_par["daymax"][:, np.newaxis]

        # yi = redshift simulated values
        # requested to avoid interpolation problems near boundaries
        yi = np.round(gen_par["z"], 4)
        # yi = gen_par['z']

        # p = phases of LC points = xi/(1.+z)
        p = xi / (1.0 + yi[:, np.newaxis])
        yi_arr = np.ones_like(p) * yi[:, np.newaxis]

        # get fluxes
        pts = (p, yi_arr)
        fluxes_obs = self.reference_lc.flux[band](pts)

        # Fisher components estimation

        d_flux = {}

        # loop on Fisher parameters
        for val in self.param__fisher:
            d_flux[val] = self.reference_lc.param[band][val](pts)

        # Fisher matrix components estimation
        # loop on SN parameters (x0,x1,color)
        # estimate: dF/dxi*dF/dxj
        derivative_for__fisher = {}
        for ia, vala in enumerate(self.param__fisher):
            for jb, valb in enumerate(self.param__fisher):
                if jb >= ia:
                    derivative_for__fisher[vala + valb] = d_flux[vala] * d_flux[valb]

        flag = self.get_flag(sel_obs, gen_par, fluxes_obs, band, p)
        flag_idx = np.argwhere(flag)

        # now apply the flag to select LC points - masked arrays
        fluxes = self.marray(fluxes_obs, flag)
        phases = self.marray(p, flag)
        z_vals = gen_par["z"][flag_idx[:, 0]]
        daymax_vals = gen_par["daymax"][flag_idx[:, 0]]
        fisher__mat = {}
        for key, vals in derivative_for__fisher.items():
            fisher__mat[key] = self.marray(vals, flag)

        ndata = len(fluxes[~fluxes.mask])

        if ndata <= 0:
            return pd.DataFrame()

        nvals = len(phases)
        # masked - tile arrays

        cols = [
            self.mjd_col,
            self.season_col,
            self.m5_col,
            self.exptime_col,
            self.nexp_col,
            self.seeing_col,
            "night",
        ]

        # take columns common with obs cols
        colcoms = set(cols).intersection(sel_obs.dtype.names)

        masked_tile = {}
        for vv in list(colcoms):
            masked_tile[vv] = self.tarray(sel_obs[vv], nvals, flag)

        # mag_obs = np.ma.array(mag_obs, mask=~flag)

        # put data in a pandas df
        lc = pd.DataFrame()

        lc["flux"] = fluxes[~fluxes.mask]
        lc["mag"] = -2.5 * np.log10(lc["flux"] / 3631.0)
        lc["phase"] = phases[~phases.mask]
        lc["band"] = ["LSST::" + band] * len(lc)
        lc["zp"] = self.reference_lc.zp[band]
        lc["zp"] = 2.5 * np.log10(3631)
        lc["zpsys"] = "ab"
        lc["z"] = z_vals
        lc["daymax"] = daymax_vals

        for key, vals in masked_tile.items():
            lc[key] = vals[~vals.mask]

        lc["time"] = lc[self.mjd_col]
        lc.loc[:, "x1"] = self.x1
        lc.loc[:, "color"] = self.color

        # estimate errors
        lc["snr_m5"] = m52snr(lc["mag"], lc["fiveSigmaDepth"])
        lc["fluxerr_photo"] = lc["flux"] / lc["snr_m5"]
        lc["magerr"] = (2.5 / np.log(10.0)) / lc["snr_m5"]

        # Fisher matrix components
        for key, vals in fisher__mat.items():
            lc.loc[:, "F_{}".format(key)] = vals[~vals.mask] / (lc["fluxerr_photo"].values ** 2)

        lc.loc[:, "n_aft"] = (np.sign(lc["phase"]) == 1) & (lc["snr_m5"] >= self.snr_min)
        lc.loc[:, "n_bef"] = (np.sign(lc["phase"]) == -1) & (lc["snr_m5"] >= self.snr_min)

        lc.loc[:, "n_phmin"] = lc["phase"] <= -5.0
        lc.loc[:, "n_phmax"] = lc["phase"] >= 20

        # if len(lc) > 0.:
        #    lc = self.dust_corrections(lc, ebvof_mw)

        # remove lc points with no detectable flux
        idx = lc["mag"] < lc["fiveSigmaDepth"] + 0.5
        lc_flux = lc[idx]

        return lc_flux

    def get_flag(self, sel_obs, gen_par, fluxes_obs, band, p):
        """Method to flag events corresponding to selection criteria

        Parameters
        ---------------
        sel_obs : `np.ndarray``
          observations
        gen_par : `np.ndarray`
           simulation parameters
        fluxes_obs : `np.ndarray`
           LC fluxes
        band : `str`
          filter
        p : `np.ndarray``
          LC phases
        """
        # remove LC points outside the restframe phase range
        min_rf_phase = gen_par["minRFphase"][:, np.newaxis]
        max_rf_phase = gen_par["maxRFphase"][:, np.newaxis]
        flag = (p >= min_rf_phase) & (p <= max_rf_phase)

        # remove LC points outside the (blue-red) range
        mean_restframe_wavelength = np.array([self.reference_lc.mean_wavelength[band]] * len(sel_obs))
        mean_restframe_wavelength = np.tile(mean_restframe_wavelength, (len(gen_par), 1)) / (
            1.0 + gen_par["z"][:, np.newaxis]
        )
        # flag &= (mean_restframe_wavelength > 0.) & (
        #    mean_restframe_wavelength < 1000000.)
        flag &= (mean_restframe_wavelength > self.blue_cutoff) & (mean_restframe_wavelength < self.red_cutoff)

        # remove points with neg flux
        flag &= fluxes_obs > 1.0e-10

        return flag

    def marray(self, val, flag):
        """method to estimate a masked array

        Parameters
        ---------------
        val : `np.ndarray`
        flag : `np.ndarray`
            mask for values

        Returns
        ----------
        masked_array : `np.ma.array`
            the masked array

        """
        return np.ma.array(val, mask=~flag)

    def tarray(self, arr, nvals, flag):
        """method to estimate a masked tiled array

        Parameters
        ---------------
        val : `np.ndarray`
        nvals : `int`
          number of tiles
        flag : np.ndarray`

        Returns
        ----------
        masked_tiles : `np.ma.array`
            the masked tiled array
        """

        vv = np.ma.array(np.tile(arr, (nvals, 1)), mask=~flag)
        return vv


def load_sne_cached(gamma_name="gamma_WFD.hdf5"):
    """Load up the SNe lightcurve files with a simple function
    that caches the result so each metric
    doesn't need to load it on it's own
    """
    need_data = True
    if hasattr(load_sne_cached, "data"):
        # Only return data if the current gamma file matches
        # the loaded gamma file
        if load_sne_cached.gammaName == gamma_name:
            need_data = False
    if need_data:
        load_sne_cached.data = LoadReference(gamma_name=gamma_name).ref
        load_sne_cached.gammaName = gamma_name
    return load_sne_cached.data


class LoadReference:
    """class to load template files requested for LCFast
    These files should be stored in a reference_files
    The files should be part of the rubin_sim_data/maf/SNe_dat files.
    They are also available for download from
    https://me.lsst.eu/gris/DESC_SN_pipeline/.

    Parameters
    ---------------
    templateDir : `str`, optional
        where to put the files (default: reference_files)
    gammaName : `str`, optional
        The name of the gamma file (for photometric zeropoint and
        SNR calculation information).
        Note that these are similar to, but not necessarily the same as,
        the current expected
        LSST zeropoints and SNR gamma values.
    """

    def __init__(
        self,
        template_dir=None,
        gamma_name="gamma_WFD.hdf5",
    ):
        if template_dir is None:
            sims_maf_contrib_dir = get_data_dir()
            template_dir = os.path.join(sims_maf_contrib_dir, "maf/SNe_data")

        x1_colors = [(-2.0, 0.2), (0.0, 0.0)]

        lc_reference = {}

        list_files = [gamma_name]
        for j in range(len(x1_colors)):
            x1 = x1_colors[j][0]
            color = x1_colors[j][1]
            fname = "LC_{}_{}_380.0_800.0_ebvofMW_0.0_vstack.hdf5".format(x1, color)
            list_files += [fname]

        # gamma_reference
        self.gamma_reference = os.path.join(template_dir, gamma_name)

        # print('Loading reference files')

        resultdict = {}

        for j in range(len(x1_colors)):
            x1 = x1_colors[j][0]
            color = x1_colors[j][1]
            fname = "{}/LC_{}_{}_380.0_800.0_ebvofMW_0.0_vstack.hdf5".format(template_dir, x1, color)
            resultdict[j] = GetReference(fname, self.gamma_reference)

        for j in range(len(x1_colors)):
            if resultdict[j] is not None:
                lc_reference[x1_colors[j]] = resultdict[j]

        self.ref = lc_reference


class GetReference:
    """Class to load reference data used for the fast SN simulator

    Parameters
    ----------------
    lcName : `str`
        name of the reference file to load (lc)
    gammaName : `str`
        name of the reference file to load (gamma)
    tel_par : `dict`
        telescope parameters
    param_Fisher : `list` [`str`], optional
        list of SN parameter for Fisher estimation to consider
        (default: ['x0', 'x1', 'color', 'daymax'])

    Notes
    -----
    After instantiations, the following dict can be accessed:
    mag_to_flux_e_sec : Interp1D of mag to flux(e.sec-1)  conversion
    flux : dict of RegularGridInterpolator of fluxes
    (key: filters, (x,y)=(phase, z), result=flux)
    fluxerr : dict of RegularGridInterpolator of flux errors
    (key: filters, (x,y)=(phase, z), result=fluxerr)
    param : dict of dict of RegularGridInterpolator of flux derivatives
    wrt SN parameters
    (key: filters plus param_Fisher parameters; (x,y)=(phase, z),
    result=flux derivatives)
    gamma : dict of RegularGridInterpolator of gamma values (key: filters)
    """

    def __init__(self, lc_name, gamma_name, param__fisher=["x0", "x1", "color", "daymax"]):
        # Load the file - lc reference
        if not os.path.exists(lc_name):
            raise FileExistsError(f"{lc_name} does not exist - NSN metric cannot run")

        # Load the file - gamma values
        if not os.path.exists(gamma_name):
            raise FileExistsError("gamma file {} does not exist - NSN metric cannot run")

        lc_ref_tot = Table.from_pandas(pd.read_hdf(lc_name))
        idx = lc_ref_tot["z"] > 0.005
        lc_ref_tot = np.copy(lc_ref_tot[idx])

        # Load references needed for the following
        self.lc_ref = {}
        self.gamma_ref = {}
        self.gamma = {}
        self.m5_ref = {}
        self.mag_to_flux_e_sec = {}
        self.mag_to_flux = {}
        self.zp = {}
        self.mean_wavelength = {}

        self.flux = {}
        self.fluxerr = {}
        self.param = {}

        bands = np.unique(lc_ref_tot["band"])

        method = "linear"

        # for each band: load data to be used for interpolation
        for band in bands:
            idx = lc_ref_tot["band"] == band
            lc_sel = Table(lc_ref_tot[idx])

            lc_sel["z"] = lc_sel["z"].data.round(decimals=2)
            lc_sel["phase"] = lc_sel["phase"].data.round(decimals=1)

            """
               select phases between -20 and 50 only
            """
            idx = lc_sel["phase"] < 60.0
            idx &= lc_sel["phase"] > -20.0
            lc_sel = lc_sel[idx]

            # these reference data will be used for griddata interp.
            self.lc_ref[band] = lc_sel
            self.gamma_ref[band] = lc_sel["gamma"][0]
            self.m5_ref[band] = np.unique(lc_sel["m5"])[0]

            # Another interpolator, faster than griddata:
            # regulargridinterpolator

            # Fluxes and errors
            zmin, zmax, zstep, nz = self.lim_vals(lc_sel, "z")
            phamin, phamax, phastep, npha = self.lim_vals(lc_sel, "phase")

            zv = np.linspace(zmin, zmax, nz)
            phav = np.linspace(phamin, phamax, npha)

            print("Loading ", lc_name, band, len(lc_sel), npha, nz)
            index = np.lexsort((lc_sel["z"], lc_sel["phase"]))
            flux = np.reshape(lc_sel[index]["flux"], (npha, nz))
            fluxerr = np.reshape(lc_sel[index]["fluxerr"], (npha, nz))

            self.flux[band] = RegularGridInterpolator(
                (phav, zv), flux, method=method, bounds_error=False, fill_value=-1.0
            )
            self.fluxerr[band] = RegularGridInterpolator(
                (phav, zv), fluxerr, method=method, bounds_error=False, fill_value=-1.0
            )

            # Flux derivatives
            self.param[band] = {}
            for par in param__fisher:
                valpar = np.reshape(lc_sel[index]["d{}".format(par)], (npha, nz))
                self.param[band][par] = RegularGridInterpolator(
                    (phav, zv),
                    valpar,
                    method=method,
                    bounds_error=False,
                    fill_value=0.0,
                )

            # gamma estimator

            rec = Table.read(gamma_name, path="gamma_{}".format(band))
            self.zp = rec.meta["zp"]
            self.mean_wavelength = rec.meta["mean_wavelength"]

            rec["mag"] = rec["mag"].data.round(decimals=4)
            rec["single_exptime"] = rec["single_exptime"].data.round(decimals=4)

            magmin, magmax, magstep, nmag = self.lim_vals(rec, "mag")
            expmin, expmax, expstep, nexpo = self.lim_vals(rec, "single_exptime")
            nexpmin, nexpmax, nexpstep, nnexp = self.lim_vals(rec, "nexp")
            mag = np.linspace(magmin, magmax, nmag)
            exp = np.linspace(expmin, expmax, nexpo)
            nexp = np.linspace(nexpmin, nexpmax, nnexp)

            index = np.lexsort((rec["nexp"], np.round(rec["single_exptime"], 4), rec["mag"]))
            gammab = np.reshape(rec[index]["gamma"], (nmag, nexpo, nnexp))
            fluxb = np.reshape(rec[index]["flux_e_sec"], (nmag, nexpo, nnexp))
            self.gamma[band] = RegularGridInterpolator(
                (mag, exp, nexp),
                gammab,
                method="linear",
                bounds_error=False,
                fill_value=0.0,
            )

            self.mag_to_flux[band] = RegularGridInterpolator(
                (mag, exp, nexp),
                fluxb,
                method="linear",
                bounds_error=False,
                fill_value=0.0,
            )

    def lim_vals(self, lc, field):
        """Get unique values of a field in  a table

        Parameters
        ----------
        lc : `astropy.Table`
            astropy Table (here probably a LC)
        field : `str`
            name of the field of interest

        Returns
        -------
        vmin : `float`
            min value of the field
        vmax : `float`
            max value of the field
        vstep : `float`
            step value for this field (median)
        nvals : `int`
            number of unique values
        """

        lc.sort(field)
        vals = np.unique(lc[field].data.round(decimals=4))
        # print(vals)
        vmin = np.min(vals)
        vmax = np.max(vals)
        vstep = np.median(vals[1:] - vals[:-1])

        return vmin, vmax, vstep, len(vals)


class SnRate:
    r"""Estimate production rates of typeIa SN
    Available rates: Ripoche, Perrett, Dilday

    Parameters
    ----------
    rate :  `str`, optional
        type of rate chosen (Ripoche, Perrett, Dilday)
    H0 : `float`, optional
        Hubble constant value :math:`H_{0}`
    Om0 : `float`, optional
        matter density value :math:`\Omega_{0}`
    min_rf_phase : `float`, optional
        min rest-frame phase
    max_rf_phase : `float`, optional
        max rest-frame phase
    """

    def __init__(self, rate="Perrett", h0=70, om0=0.25, min_rf_phase=-15.0, max_rf_phase=30.0):
        self.astropy_cosmo = FlatLambdaCDM(H0=h0, Om0=om0)
        self.rate = rate
        self.min_rf_phase = min_rf_phase
        self.max_rf_phase = max_rf_phase

    def __call__(
        self,
        zmin=0.1,
        zmax=0.2,
        dz=0.01,
        survey_area=9.6,
        bins=None,
        account_for_edges=False,
        duration=140.0,
        duration_z=None,
    ):
        """
        Parameters
        ----------------
        zmin : `float`, optional
            minimal redshift (default : 0.1)
        zmax : `float`, optional
            max redshift (default : 0.2)
        dz : `float`, optional
            redshift bin (default : 0.001)
        survey_area : `float`, optional
            area of the survey (:math:`deg^{2}`)
        bins : `list` [`float`], optional
            redshift bins
        account_for_edges : `bool`
            to account for season edges.
            If true, duration of the survey will be reduced by
            (1+z)*(maf_rf_phase-min_rf_phase)/365.25
        duration : `float`, optional
            survey duration (in days)
        duration_z : `list` [`float`], optional
            survey duration (as a function of z)

        Returns
        -----------
        zz : `list` [`float`]
            redshift values
        rate : `list` [`float`]
            production rate
        err_rate : `list` [`float`]
            production rate error
        nsn : `list` [`float`]
            number of SN
        err_nsn : `list` [`float`]
            error on the number of SN
        """

        if bins is None:
            thebins = np.arange(zmin, zmax + dz, dz)
            zz = 0.5 * (thebins[1:] + thebins[:-1])
        else:
            zz = bins
            thebins = bins

        rate, err_rate = self.sn_rate(zz)

        area = survey_area / steradian2_sqdeg
        # or area= self.survey_area/41253.

        dvol = norm * self.astropy_cosmo.comoving_volume(thebins).value

        dvol = dvol[1:] - dvol[:-1]

        if account_for_edges:
            margin = (1.0 + zz) * (self.max_rf_phase - self.min_rf_phase) / 365.25
            effective_duration = duration / 365.25 - margin
            effective_duration[effective_duration <= 0.0] = 0.0
        else:
            # duration in days!
            effective_duration = duration / 365.25
            if duration_z is not None:
                effective_duration = duration_z(zz) / 365.25

        normz = 1.0 + zz
        nsn = rate * area * dvol * effective_duration / normz
        err_nsn = err_rate * area * dvol * effective_duration / normz

        return zz, rate, err_rate, nsn, err_nsn

    def ripoche_rate(self, z):
        """The SNLS SNIa rate according to the (unpublished)
        Ripoche et al study.

        Parameters
        --------------
        z : `float`
            redshift

        Returns
        ----------
        rate : `float`
        error_rate : `float`
        """
        rate = 1.53e-4 * 0.343
        expn = 2.14
        my_z = np.copy(z)
        my_z[my_z > 1.0] = 1.0
        rate_sn = rate * np.power((1 + my_z) / 1.5, expn)
        return rate_sn, 0.2 * rate_sn

    def perrett_rate(self, z):
        """The SNLS SNIa rate according to (Perrett et al, 201?)

        Parameters
        --------------
        z : `float`
            redshift

        Returns
        ----------
        rate : `float`
        error_rate : `float`
        """
        rate = 0.17e-4
        expn = 2.11
        err_rate = 0.03e-4
        err_expn = 0.28
        my_z = np.copy(z)
        rate_sn = rate * np.power(1 + my_z, expn)
        err_rate_sn = np.power(1 + my_z, 2.0 * expn) * np.power(err_rate, 2.0)
        err_rate_sn += np.power(rate_sn * np.log(1 + my_z) * err_expn, 2.0)

        return rate_sn, np.power(err_rate_sn, 0.5)

    def dilday_rate(self, z):
        """The Dilday rate according to

        Parameters
        --------------
        z : `float`
            redshift

        Returns
        ----------
        rate : `float`
        error_rate : `float`
        """

        rate = 2.6e-5
        expn = 1.5
        # err_rate = 0.01
        err_expn = 0.6
        my_z = np.copy(z)
        my_z[my_z > 1.0] = 1.0
        rate_sn = rate * np.power(1 + my_z, expn)
        err_rate_sn = rate_sn * np.log(1 + my_z) * err_expn
        return rate_sn, err_rate_sn

    def sn_rate(self, z):
        """SN rate estimation

        Parameters
        --------------
        z : `float`
            redshift

        Returns
        ----------
        rate : `float`
        error_rate : `float`
        """
        if self.rate == "Ripoche":
            return self.ripoche_rate(z)
        if self.rate == "Perrett":
            return self.perrett_rate(z)
        if self.rate == "Dilday":
            return self.dilday_rate(z)


class CovColor:
    """
    class to estimate CovColor from lc using Fisher matrix element

    Parameters
    ---------------
    lc : `pandas.DataFrame`
        lc to process. Should contain the Fisher matrix components
        ie the sum of the derivative of the fluxes wrt SN parameters
    """

    def __init__(self, lc):
        self.cov_colorcolor = self.var_color(lc)

    def var_color(self, lc):
        """
        Method to estimate the variance color from matrix element

        Parameters
        --------------
        lc : `pandas.DataFrame`
            data to process containing the derivative of the
            flux with respect to SN parameters

        Returns
        ----------
        Cov_colorcolor : `float`
        """
        a1 = lc["F_x0x0"]
        a2 = lc["F_x0x1"]
        a3 = lc["F_x0daymax"]
        a4 = lc["F_x0color"]

        b1 = a2
        b2 = lc["F_x1x1"]
        b3 = lc["F_x1daymax"]
        b4 = lc["F_x1color"]

        c1 = a3
        c2 = b3
        c3 = lc["F_daymaxdaymax"]
        c4 = lc["F_daymaxcolor"]

        d1 = a4
        d2 = b4
        d3 = c4
        d4 = lc["F_colorcolor"]

        det_m = a1 * self.det(b2, b3, b4, c2, c3, c4, d2, d3, d4)
        det_m -= b1 * self.det(a2, a3, a4, c2, c3, c4, d2, d3, d4)
        det_m += c1 * self.det(a2, a3, a4, b2, b3, b4, d2, d3, d4)
        det_m -= d1 * self.det(a2, a3, a4, b2, b3, b4, c2, c3, c4)

        res = -a3 * b2 * c1 + a2 * b3 * c1 + a3 * b1 * c2 - a1 * b3 * c2 - a2 * b1 * c3 + a1 * b2 * c3

        return res / det_m

    def det(self, a1, a2, a3, b1, b2, b3, c1, c2, c3):
        """
        Method to estimate the det of a matrix from its values

        Parameters
        -------------
        Values of the matrix
        (a1 a2 a3)
        (b1 b2 b3)
        (c1 c2 c3)

        Returns
        -----------
        det value
        """
        resp = a1 * b2 * c3 + b1 * c2 * a3 + c1 * a2 * b3
        resm = a3 * b2 * c1 + b3 * c2 * a1 + c3 * a2 * b1

        return resp - resm



import os
import unittest
import warnings

import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.maf.metrics import SNCadenceMetric, SNSLMetric, SNSNRMetric
from rubin_sim.maf.utils.sn_utils import Lims, ReferenceData

m5_ref = dict(zip("ugrizy", [23.60, 24.83, 24.38, 23.92, 23.35, 22.44]))


def observations_band(day0=59000, daymin=59000, cadence=3.0, season_length=140.0, band="r"):
    # Define fake data
    names = [
        "observationStartMJD",
        "fieldRA",
        "fieldDec",
        "fiveSigmaDepth",
        "visitExposureTime",
        "numExposures",
        "visitTime",
        "season",
        "seeingFwhmEff",
        "seeingFwhmGeom",
        "pixRA",
        "pixDec",
    ]
    types = ["f8"] * len(names)
    names += ["night", "healpixID"]
    types += ["i2", "i2"]
    names += ["filter"]
    types += ["O"]

    daylast = daymin + season_length
    cadence = cadence
    dayobs = np.arange(daymin, daylast, cadence)
    npts = len(dayobs)
    data = np.zeros(npts, dtype=list(zip(names, types)))
    data["observationStartMJD"] = dayobs
    data["night"] = np.floor(data["observationStartMJD"] - day0 + 1)
    data["fiveSigmaDepth"] = m5_ref[band]
    data["visitExposureTime"] = 30.0
    data["numExposures"] = 1
    data["visitTime"] = 34
    data["filter"] = band
    data["seeingFwhmEff"] = 0.8
    data["seeingFwhmGeom"] = 0.8
    data["healpixID"] = 10
    data["pixRA"] = 0.0
    data["pixDec"] = 0.0
    return data


def observations_season(mjdmin=59000, cadence=3.0):
    bands = "grizy"
    nvisits = dict(zip(bands, [10, 20, 20, 26, 20]))
    rat = 34.0 / 3600.0 / 24.0
    shift_visits = {}
    shift_visits["g"] = 0
    shift_visits["r"] = rat * nvisits["g"]
    shift_visits["i"] = rat * nvisits["r"]
    shift_visits["z"] = rat * nvisits["i"]
    shift_visits["y"] = rat * nvisits["z"]

    # get data
    data = None
    season_length = 180
    shift = 30.0 / (3600.0 * 24)
    for band in bands:
        mjd = mjdmin + shift_visits[band]
        for i in range(nvisits[band]):
            mjd += shift
            dat = observations_band(daymin=mjd, season_length=season_length, cadence=cadence, band=band)
            if data is None:
                data = dat
            else:
                data = np.concatenate((data, dat))

    return data


def fake_data(band, season=1):
    # Define fake data
    names = [
        "observationStartMJD",
        "fieldRA",
        "fieldDec",
        "fiveSigmaDepth",
        "visitExposureTime",
        "numExposures",
        "visitTime",
        "season",
        "seeingFwhmEff",
        "seeingFwhmGeom",
        "airmass",
        "sky",
        "moonPhase",
        "pixRA",
        "pixDec",
    ]

    types = ["f8"] * len(names)
    names += ["night"]
    types += ["i2"]
    names += ["healpixID"]
    types += ["i2"]
    names += ["filter"]
    types += ["O"]

    dayobs = [
        59948.31957176,
        59959.2821412,
        59970.26134259,
        59973.25978009,
        59976.26383102,
        59988.20670139,
        59991.18412037,
        60004.1853588,
        60032.08975694,
        60045.11981481,
        60047.98747685,
        60060.02083333,
        60071.986875,
        60075.96452546,
    ]
    day0 = np.min(dayobs)
    npts = len(dayobs)
    data = np.zeros(npts, dtype=list(zip(names, types)))
    data["observationStartMJD"] = dayobs
    data["night"] = np.floor(data["observationStartMJD"] - day0 + 1)
    data["fiveSigmaDepth"] = m5_ref[band]
    data["visitExposureTime"] = 15.0
    data["numExposures"] = 2
    data["visitTime"] = 2.0 * 15.0
    data["season"] = season
    data["filter"] = band
    data["seeingFwhmEff"] = 0.0
    data["seeingFwhmGeom"] = 0.0
    data["airmass"] = 1.2
    data["sky"] = 20.0
    data["moonPhase"] = 0.5
    data["pixRA"] = 0.0
    data["pixDec"] = 0.0
    data["healpixID"] = 1

    return data


class TestSNmetrics(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir(os.path.join(get_data_dir(), "maf")):
            self.skipTest("Skipping SN tests because running unit tests without full rubin_sim.data.")

    @unittest.skip("The SNCadenceMetric is not used")
    def test_sn_cadence_metric(self):
        """Test the SN cadence metric"""

        # Load up the files from sims_maf_contrib if possible
        sims_maf_contrib_dir = os.path.join(get_data_dir(), "maf")
        if sims_maf_contrib_dir is not None:
            # Load required SN info to run the metric
            band = "r"
            SNR = dict(zip("griz", [30.0, 40.0, 30.0, 20.0]))  # SNR for WFD
            mag_range = [21.0, 25.5]  # WFD mag range
            dt_range = [0.5, 30.0]  # WFD dt range
            li_files = [os.path.join(sims_maf_contrib_dir, "Li_SNCosmo_-2.0_0.2.npy")]
            mag_to_flux_files = [os.path.join(sims_maf_contrib_dir, "Mag_to_Flux_SNCosmo.npy")]
            lim_sn = Lims(
                li_files,
                mag_to_flux_files,
                band,
                SNR[band],
                mag_range=mag_range,
                dt_range=dt_range,
            )

            # Define fake data
            names = [
                "observationStartMJD",
                "fieldRA",
                "fieldDec",
                "fiveSigmaDepth",
                "visitExposureTime",
                "numExposures",
                "visitTime",
            ]
            types = ["f8"] * len(names)
            names += ["night"]
            types += ["i2"]
            names += ["filter"]
            types += ["O"]

            day0 = 59000
            daylast = day0 + 250
            cadence = 5
            dayobs = np.arange(day0, daylast, cadence)
            npts = len(dayobs)
            data = np.zeros(npts, dtype=list(zip(names, types)))
            data["observationStartMJD"] = dayobs
            data["night"] = np.floor(data["observationStartMJD"] - day0)
            data["fiveSigmaDepth"] = m5_ref[band]
            data["visitExposureTime"] = 15.0
            data["numExposures"] = 2
            data["visitTime"] = 2.0 * 15.0
            data["filter"] = band

            # Run the metric with these fake data
            slice_point = {"nside": 64, "ebv": 0}
            metric = SNCadenceMetric(lim_sn=lim_sn, coadd=False)
            result = metric.run(data, slice_point)

            # And the result should be...
            result_ref = 0.3743514

            assert np.abs(result - result_ref) < 1.0e-5

        else:
            warnings.warn("skipping SN test because no rubin_sim.data set")

    @unittest.skip("This metric is not used")
    def test_snsnr_metric(self):
        """Test the SN SNR metric"""

        sims_maf_contrib_dir = os.path.join(get_data_dir(), "maf")
        if sims_maf_contrib_dir is not None:
            # Load required SN info to run the metric
            band = "r"
            z = 0.3
            season = 1.0
            li_files = [os.path.join(sims_maf_contrib_dir, "Li_SNCosmo_-2.0_0.2.npy")]
            mag_to_flux_files = [os.path.join(sims_maf_contrib_dir, "Mag_to_Flux_SNCosmo.npy")]

            names_ref = ["SNCosmo"]
            coadd = False

            lim_sn = ReferenceData(li_files, mag_to_flux_files, band, z)

            # Define fake data
            names = [
                "observationStartMJD",
                "fieldRA",
                "fieldDec",
                "fiveSigmaDepth",
                "visitExposureTime",
                "numExposures",
                "visitTime",
                "season",
            ]
            types = ["f8"] * len(names)
            names += ["night"]
            types += ["i2"]
            names += ["filter"]
            types += ["O"]

            dayobs = [
                59948.31957176,
                59959.2821412,
                59970.26134259,
                59973.25978009,
                59976.26383102,
                59988.20670139,
                59991.18412037,
                60004.1853588,
                60032.08975694,
                60045.11981481,
                60047.98747685,
                60060.02083333,
                60071.986875,
                60075.96452546,
            ]
            day0 = np.min(dayobs)
            npts = len(dayobs)
            data = np.zeros(npts, dtype=list(zip(names, types)))

            data["observationStartMJD"] = dayobs
            data["night"] = np.floor(data["observationStartMJD"] - day0)
            data["fiveSigmaDepth"] = m5_ref[band]
            data["visitExposureTime"] = 15.0
            data["numExposures"] = 2
            data["visitTime"] = 2.0 * 15.0
            data["season"] = season
            data["filter"] = band

            # Run the metric with these fake data
            slice_point = {"nside": 64, "ebv": 0.0}
            metric = SNSNRMetric(lim_sn=lim_sn, coadd=coadd, names_ref=names_ref, season=season, z=z)

            result = metric.run(data, slice_point)

            # And the result should be...
            result_ref = 0.4830508474576271

            assert np.abs(result - result_ref) < 1.0e-5
        else:
            warnings.warn("skipping SN test because no rubin_sim.data set")

    def test_snsl_metric(self):
        """Test the SN SL metric"""

        # load some fake data
        data = None
        bands = "griz"
        cadence = dict(zip(bands, [2, 1, 2, 1]))
        for band in bands:
            for i in range(cadence[band]):
                fakes = fake_data(band)
                if data is None:
                    data = fakes
                else:
                    data = np.concatenate((data, fakes))

        # metric instance

        metric = SNSLMetric(
            nfilters_min=4,
            min_season_obs=5,
            m5mins={"u": 22.7, "g": 24.1, "r": 23.7, "i": 23.1, "z": 22.2, "y": 21.4},
        )

        # run the metric
        n_sl = metric.run(data, slice_point={"nside": 64, "ra": 0.0, "ebv": 0.0})

        # and the result should be
        # Changing the reference value because we have new coadd and mag limits
        # Change again to switch to per-season calc rather than average
        # Change again to reflect updated calculation
        # (due to counting year = 365.25 days, not 360)
        n_sl_ref = 1.4569195613987307e-06
        assert np.abs(n_sl - n_sl_ref) < 1.0e-8


if __name__ == "__main__":
    unittest.main()



__all__ = ("altazHealpix", "altazLambert")

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict
from .common import filter_list


def basicSetup(metric_name, colmap=None, nside=64):
    if colmap is None:
        colmap = col_map_dict()

    slicer = slicers.HealpixSlicer(
        nside=nside,
        lat_col=colmap["alt"],
        lon_col=colmap["az"],
        lat_lon_deg=colmap["raDecDeg"],
        use_cache=False,
    )
    metric = metrics.CountMetric(colmap["mjd"], metric_name=metric_name)

    return colmap, slicer, metric


def altazHealpix(
    colmap=None,
    run_name="run name",
    extraSql=None,
    extraInfoLabel=None,
    metric_name="NVisits Alt/Az",
):
    """
    Generate a set of metrics measuring the number visits as a function
    of alt/az plotted on a HealpixSkyMap.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").
    metric_name : `str`, optional
        Unique name to assign to metric

    Returns
    -------
    metric_bundleDict : `dict` of {`str`: `maf.MetricBundle`}
    """

    colmap, slicer, metric = basicSetup(metric_name=metric_name, colmap=colmap)

    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    bundleList = []

    plotDict = {"rot": (90, 90, 90), "flip": "geo"}
    plotFunc = plots.HealpixSkyMap()

    for f in filterlist:
        if f == "all":
            subgroup = "All Observations"
        else:
            subgroup = "Per filter"
        displayDict = {
            "group": "Alt/Az",
            "order": orders[f],
            "subgroup": subgroup,
            "caption": "Pointing History on the alt-az sky (zenith center) for filter %s" % f,
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            plot_dict=plotDict,
            run_name=run_name,
            info_label=info_label[f],
            plot_funcs=[plotFunc],
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def altazLambert(
    colmap=None,
    runName="opsim",
    extraSql=None,
    extraInfoLabel=None,
    metric_name="Nvisits as function of Alt/Az",
):
    """
    Generate a set of metrics measuring the number visits as a
    function of alt/az plotted on a LambertSkyMap.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").
    metric_name : `str`, optional
        Unique name to assign to metric

    Returns
    -------
    metric_bundleDict : `dict` of {`str`: `maf.MetricBundle`}
    """

    colmap, slicer, metric = basicSetup(metric_name=metric_name, colmap=colmap)

    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    bundleList = []

    plotFunc = plots.LambertSkyMap()

    for f in filterlist:
        if f == "all":
            subgroup = "All Observations"
        else:
            subgroup = "Per filter"
        displayDict = {
            "group": "Alt/Az",
            "order": orders[f],
            "subgroup": subgroup,
            "caption": "Alt/Az pointing distribution for filter %s" % f,
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            run_name=runName,
            info_label=info_label[f],
            plot_funcs=[plotFunc],
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



__all__ = (
    "NormAirmassStacker",
    "ParallaxFactorStacker",
    "HourAngleStacker",
    "ZenithDistStacker",
    "ParallacticAngleStacker",
    "DcrStacker",
    "FiveSigmaStacker",
    "SaturationStacker",
    "OverheadStacker",
)

import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import GCRS, SkyCoord
from astropy.time import Time
from rubin_scheduler.utils import Site, m5_flat_sed

from rubin_sim.maf.utils import load_inst_zeropoints

from .base_stacker import BaseStacker
from .date_stackers import DayObsMJDStacker


class SaturationStacker(BaseStacker):
    """Adds a calculated point-source saturation limit for each visit.

    Assumes Gaussian PSF.

    Parameters
    ----------
    pixscale : `float`, optional
        Arcsec per pixel.
    saturation_e : `float`, optional
        The saturation level in electrons.
    zeropoints : dict-like, optional
        The zeropoints for the telescope.
        Keys should be str with filter names, values in mags.
        Default of None, will use Rubin calculated zeropoints.
    km : dict-like, optional
        Atmospheric extinction values.
        Keys should be str with filter names.
        If None, will use Rubin calculated atmospheric extinction values.
    """

    cols_added = ["saturation_mag"]

    def __init__(
        self,
        seeing_col="seeingFwhmEff",
        skybrightness_col="skyBrightness",
        exptime_col="visitExposureTime",
        nexp_col="numExposures",
        filter_col="filter",
        airmass_col="airmass",
        saturation_e=150e3,
        zeropoints=None,
        km=None,
        pixscale=0.2,
    ):
        self.units = ["mag"]
        self.cols_req = [
            seeing_col,
            skybrightness_col,
            exptime_col,
            nexp_col,
            filter_col,
            airmass_col,
        ]
        self.seeing_col = seeing_col
        self.skybrightness_col = skybrightness_col
        self.exptime_col = exptime_col
        self.nexp_col = nexp_col
        self.filter_col = filter_col
        self.airmass_col = airmass_col
        self.saturation_e = saturation_e
        self.pixscale = pixscale
        self.zeropoints = zeropoints
        self.km = km

    def _run(self, sim_data, cols_present=False):
        if self.zeropoints is None:
            zp_inst, k_atm = load_inst_zeropoints()
            self.zeropoints = zp_inst
        if self.km is None:
            self.km = k_atm
        for filtername in np.unique(sim_data[self.filter_col]):
            in_filt = np.where(sim_data[self.filter_col] == filtername)[0]
            # Calculate the length of the on-sky time per EXPOSURE
            exptime = sim_data[self.exptime_col][in_filt] / sim_data[self.nexp_col][in_filt]
            # Calculate sky counts per pixel per second
            # from skybrightness + zeropoint (e/1s)
            sky_counts = (
                10.0 ** (0.4 * (self.zeropoints[filtername] - sim_data[self.skybrightness_col][in_filt]))
                * self.pixscale**2
            )
            # Total sky counts in each exposure
            sky_counts = sky_counts * exptime
            # The counts available to the source (at peak) in each exposure is
            # the difference between saturation and sky
            remaining_counts_peak = self.saturation_e - sky_counts
            # Now to figure out how many counts there would be total, if there
            # are that many in the peak
            sigma = sim_data[self.seeing_col][in_filt] / 2.354
            source_counts = remaining_counts_peak * 2.0 * np.pi * (sigma / self.pixscale) ** 2
            # source counts = counts per exposure (expTimeCol / nexp)
            # Translate to counts per second, to apply zeropoint
            count_rate = source_counts / exptime
            sim_data["saturation_mag"][in_filt] = -2.5 * np.log10(count_rate) + self.zeropoints[filtername]
            # Airmass correction
            sim_data["saturation_mag"][in_filt] -= self.km[filtername] * (
                sim_data[self.airmass_col][in_filt] - 1.0
            )
            # Explicitly make sure if sky has saturated we return NaN
            sim_data["saturation_mag"][np.where(remaining_counts_peak < 0)] = np.nan

        return sim_data


class NoteRootStacker(BaseStacker):
    """Strip off things after a comma in the scheduler_note"""

    cols_added = ["scheduler_note_root"]

    def __init__(self, note_col="scheduler_note"):
        self.units = ["mag"]
        self.cols_req = [note_col]
        self.note_col = note_col
        self.cols_added_dtypes = ["<U50"]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is fine.
            return sim_data
        new_note = [note.split(",")[0] for note in sim_data[self.note_col]]
        sim_data["scheduler_note_root"] = new_note
        return sim_data


class FiveSigmaStacker(BaseStacker):
    """Calculate the 5-sigma limiting depth for a point source in the given
    conditions.

    This is generally not needed, unless the m5 parameters have been updated
    or m5 was not previously calculated.

    Parameters
    ----------
    airmass_col : `str`, optional
        Name of the airmass column in the data.
    seeing_col : `str`, optional
        Name of the seeing column in the data.
        (FWHM of the single-Gaussian PSF)
    skybrightness_col : `str`, optional
        Name of the skybrightness column in the data.
    filter_col : `str`, optional
        Name of the filter bandpass column in the data.
    exptime_col : `str`, optional
        Name of the on-sky exposure time column in the data.
    """

    cols_added = ["m5_simsUtils"]

    def __init__(
        self,
        airmass_col="airmass",
        seeing_col="seeingFwhmEff",
        skybrightness_col="skyBrightness",
        filter_col="filter",
        exptime_col="visitExposureTime",
    ):
        self.units = ["mag"]
        self.cols_req = [
            airmass_col,
            seeing_col,
            skybrightness_col,
            filter_col,
            exptime_col,
        ]
        self.airmass_col = airmass_col
        self.seeing_col = seeing_col
        self.skybrightness_col = skybrightness_col
        self.filter_col = filter_col
        self.exptime_col = exptime_col

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is fine.
            return sim_data
        filts = np.unique(sim_data[self.filter_col])
        for filtername in filts:
            infilt = np.where(sim_data[self.filter_col] == filtername)
            sim_data["m5_simsUtils"][infilt] = m5_flat_sed(
                filtername,
                sim_data[infilt][self.skybrightness_col],
                sim_data[infilt][self.seeing_col],
                sim_data[infilt][self.exptime_col],
                sim_data[infilt][self.airmass_col],
            )
        return sim_data


class NormAirmassStacker(BaseStacker):
    """Adds a calculated normairmass for each pointing.

    The normalized airmass is the airmass divided by the minimum airmass
    achievable at each pointing (which is defined by the declination of
    the field).

    Parameters
    ----------
    airmass_col : `str`, optional
        The name of the airmass column in the data.
    dec_col : `str`, optional
        The name of the declination column in the data.
    degrees : `bool`, optional
        If True, angle columns are assumed to be in degrees and returned in
        degrees. If False, uses and calculates radians.
    telescope_lat : `float`, optional
        The latitude of the telescope, in degrees.
    """

    cols_added = ["normairmass"]

    def __init__(
        self,
        airmass_col="airmass",
        dec_col="fieldDec",
        degrees=True,
        telescope_lat=-30.2446388,
    ):
        self.units = ["X / Xmin"]
        self.cols_req = [airmass_col, dec_col]
        self.airmass_col = airmass_col
        self.dec_col = dec_col
        self.telescope_lat = telescope_lat
        self.degrees = degrees

    def _run(self, sim_data, cols_present=False):
        # Run method is required to calculate column.
        # Driver runs getColInfo to know what columns are needed from db &
        # which are calculated,  then gets data from db and then calculates
        # additional columns (via run methods here).
        if cols_present:
            # Column already present in data; assume it is correct
            # and does not need recalculating.
            return sim_data
        dec = sim_data[self.dec_col]
        if self.degrees:
            dec = np.radians(dec)
        min_z_possible = np.abs(dec - np.radians(self.telescope_lat))
        min_airmass_possible = 1.0 / np.cos(min_z_possible)
        sim_data["normairmass"] = sim_data[self.airmass_col] / min_airmass_possible
        return sim_data


class ZenithDistStacker(BaseStacker):
    """Adds a calculated zenithDistance value for each pointing.

    Parameters
    ----------
    alt_col : `str`, optional
        The name of the altitude column in the data.
    degrees : `bool`, optional
        If True, data in alt_col is in degrees, and values for zenithDistance
        will be in degrees. (Default).
        If False, data in alt_col is in radians and zenithDistance values
        will be in radians.
    """

    cols_added = ["zenithDistance"]

    def __init__(self, alt_col="altitude", degrees=True):
        self.alt_col = alt_col
        self.degrees = degrees
        if self.degrees:
            self.units = ["degrees"]
        else:
            self.unit = ["radians"]
        self.cols_req = [self.alt_col]

    def _run(self, sim_data, cols_present=False):
        """Calculate new column for zenith distance."""
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data
        if self.degrees:
            sim_data["zenithDistance"] = 90.0 - sim_data[self.alt_col]
        else:
            sim_data["zenithDistance"] = np.pi / 2.0 - sim_data[self.alt_col]
        return sim_data


class ParallaxFactorStacker(BaseStacker):
    """Add a parallax factor (in arcseconds) column for each visit.

    Parameters
    ----------
    ra_col : `str`, optional
        Name of the RA column in the data.
    dec_col : `str`, optional
        Name of the declination column in the data.
    date_col : `str`, optional
        Name of the exposure start time column in the data.
        Date should be in units of MJD.
    degrees : `bool`, optional
        If true, assumes angles are in degrees. If False, radians.
    """

    cols_added = ["ra_pi_amp", "dec_pi_amp"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        date_col="observationStartMJD",
        degrees=True,
    ):
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.date_col = date_col
        self.units = ["arcsec", "arcsec"]
        self.cols_req = [ra_col, dec_col, date_col]
        self.degrees = degrees

    def _gnomonic_project_toxy(self, ra1, dec1, r_acen, deccen):
        # ra, dec values in RADIANS
        # also used in Global Telescope Network website
        cosc = np.sin(deccen) * np.sin(dec1) + np.cos(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)
        x = np.cos(dec1) * np.sin(ra1 - r_acen) / cosc
        y = (np.cos(deccen) * np.sin(dec1) - np.sin(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)) / cosc
        return x, y

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct
            # and does not need recalculating.
            return sim_data
        ra_pi_amp = np.zeros(np.size(sim_data), dtype=[("ra_pi_amp", "float")])
        dec_pi_amp = np.zeros(np.size(sim_data), dtype=[("dec_pi_amp", "float")])
        ra = sim_data[self.ra_col]
        dec = sim_data[self.dec_col]
        if self.degrees:
            ra = np.radians(ra)
            dec = np.radians(dec)

        times = Time(sim_data[self.date_col], format="mjd")
        c = SkyCoord(ra * u.rad, dec * u.rad, obstime=times)
        geo_far = c.transform_to(GCRS)
        c_near = SkyCoord(ra * u.rad, dec * u.rad, distance=1 * u.pc, obstime=times)
        geo_near = c_near.transform_to(GCRS)

        x_geo1, y_geo1 = self._gnomonic_project_toxy(geo_near.ra.rad, geo_near.dec.rad, ra, dec)
        x_geo, y_geo = self._gnomonic_project_toxy(geo_far.ra.rad, geo_far.dec.rad, ra, dec)

        # Return ra_pi_amp and dec_pi_amp in arcseconds.
        ra_pi_amp[:] = np.degrees(x_geo1 - x_geo) * 3600.0
        dec_pi_amp[:] = np.degrees(y_geo1 - y_geo) * 3600.0
        sim_data["ra_pi_amp"] = ra_pi_amp
        sim_data["dec_pi_amp"] = dec_pi_amp
        return sim_data


class DcrStacker(BaseStacker):
    """Add columns representing the expected RA/Dec offsets expected for
    an object due to differential chromatic refraction, per visit.

    For DCR calculation, we also need zenithDistance, HA, and PA -- but these
    will be explicitly handled within this stacker so that setup is consistent
    and they run in order. If those values have already been calculated
    elsewhere, they will not be overwritten.

    Parameters
    ----------
    filter_col : `str`, optional
        The name of the column with filter names. Default 'filter'.
    altCol : `str`, optional
        Name of the column with altitude info. Default 'altitude'.
    ra_col : `str`, optional
        Name of the column with RA. Default 'fieldRA'.
    dec_col : `str`, optional
        Name of the column with Dec. Default 'fieldDec'.
    lstCol : `str`, optional
        Name of the column with local sidereal time. Default
        'observationStartLST'.
    site : `str` or `rubin_scheduler.utils.Site`, optional
        Name of the observory or a rubin_scheduler.utils.Site object.
        Default 'LSST'.
    mjdCol : `str`, optional
        Name of column with modified julian date.
        Default 'observationStartMJD'
    dcr_magnitudes : dict
        Magnitude of the DCR offset for each filter at an
        altitude/zenith distance of 45 degrees.
        Defaults u=0.07, g=0.07, r=0.50, i=0.045, z=0.042, y=0.04
        (all values should be in arcseconds).

    Returns
    -------
    data : `numpy.array`
        Returns array with additional columns 'ra_dcr_amp' and 'dec_dcr_amp'
        with the DCR offsets for each observation.  Also runs ZenithDistStacker
        and ParallacticAngleStacker.
    """

    cols_added = ["ra_dcr_amp", "dec_dcr_amp"]  # zenithDist, HA, PA

    def __init__(
        self,
        filter_col="filter",
        alt_col="altitude",
        degrees=True,
        ra_col="fieldRA",
        dec_col="fieldDec",
        lst_col="observationStartLST",
        site="LSST",
        mjd_col="observationStartMJD",
        dcr_magnitudes=None,
    ):
        self.units = ["arcsec", "arcsec"]
        if dcr_magnitudes is None:
            # DCR amplitudes are in arcseconds.
            self.dcr_magnitudes = {
                "u": 0.07,
                "g": 0.07,
                "r": 0.050,
                "i": 0.045,
                "z": 0.042,
                "y": 0.04,
            }
        else:
            self.dcr_magnitudes = dcr_magnitudes
        self.zd_col = "zenithDistance"
        self.pa_col = "PA"
        self.filter_col = filter_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        self.cols_req = [filter_col, ra_col, dec_col, alt_col, lst_col]
        #  'zenithDist', 'PA', 'HA' are additional columns required, coming
        #  from other stackers which must also be configured -- so we handle
        #  this explicitly here.
        self.zstacker = ZenithDistStacker(alt_col=alt_col, degrees=self.degrees)
        self.pastacker = ParallacticAngleStacker(
            ra_col=ra_col,
            dec_col=dec_col,
            mjd_col=mjd_col,
            degrees=self.degrees,
            lst_col=lst_col,
            site=site,
        )
        # Note that RA/Dec could be coming from a dither stacker!
        # But we will assume that coord stackers will be handled separately.

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data
        # Need to make sure the Zenith stacker gets run first Call _run method
        # because already added these columns due to 'colsAdded' line.
        sim_data = self.zstacker.run(sim_data)
        sim_data = self.pastacker.run(sim_data)
        if self.degrees:
            zenith_tan = np.tan(np.radians(sim_data[self.zd_col]))
            parallactic_angle = np.radians(sim_data[self.pa_col])
        else:
            zenith_tan = np.tan(sim_data[self.zd_col])
            parallactic_angle = sim_data[self.pa_col]
        dcr_in_ra = zenith_tan * np.sin(parallactic_angle)
        dcr_in_dec = zenith_tan * np.cos(parallactic_angle)
        for filtername in np.unique(sim_data[self.filter_col]):
            fmatch = np.where(sim_data[self.filter_col] == filtername)
            dcr_in_ra[fmatch] = self.dcr_magnitudes[filtername] * dcr_in_ra[fmatch]
            dcr_in_dec[fmatch] = self.dcr_magnitudes[filtername] * dcr_in_dec[fmatch]
        sim_data["ra_dcr_amp"] = dcr_in_ra
        sim_data["dec_dcr_amp"] = dcr_in_dec
        return sim_data


class HourAngleStacker(BaseStacker):
    """Add the Hour Angle (in decimal hours) for each observation.

    Parameters
    ----------
    lst_col : `str`, optional
        Name of the LST column in the data.
    ra_col : `str`, optional
        Name of the RA column in the data.
    degrees : `bool`, optional
        If True, assumes angles (RA and LST) are in degrees.
        If False, assumes radians.
    """

    cols_added = ["HA"]

    def __init__(self, lst_col="observationStartLST", ra_col="fieldRA", degrees=True):
        self.units = ["Hours"]
        self.cols_req = [lst_col, ra_col]
        self.lst_col = lst_col
        self.ra_col = ra_col
        self.degrees = degrees

    def _run(self, sim_data, cols_present=False):
        """HA = LST - RA"""
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data
        if len(sim_data) == 0:
            return sim_data
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            lst = np.radians(sim_data[self.lst_col])
        else:
            ra = sim_data[self.ra_col]
            lst = sim_data[self.lst_col]
        # Check that LST is reasonable
        if (np.min(lst) < 0) | (np.max(lst) > 2.0 * np.pi):
            warnings.warn("LST values are not between 0 and 2 pi")
        # Check that RA is reasonable
        if (np.min(ra) < 0) | (np.max(ra) > 2.0 * np.pi):
            warnings.warn("RA values are not between 0 and 2 pi")
        ha = lst - ra
        # Wrap the results so HA between -pi and pi
        ha = np.where(ha < -np.pi, ha + 2.0 * np.pi, ha)
        ha = np.where(ha > np.pi, ha - 2.0 * np.pi, ha)
        # Convert radians to hours
        sim_data["HA"] = ha * 12 / np.pi
        return sim_data


class ParallacticAngleStacker(BaseStacker):
    """Add the calculated parallactic angle to each visit.

    Parameters
    ----------
    ra_col : `str`, optional
        Name of the RA column in the data.
    dec_col : `str`, optional
        Name of the declination column in the data.
    degrees : `bool`, optional
        If True, assumes ra and dec in degrees and returns Parallactic Angle
        in degrees. If False, assumes and returns radians.
    mjd_col : `str`, optional
        Name of the observation MJD column in the data.
    lst_col : `str`, optional
        Name of the LST column in the data.
    site : `str` or `rubin_scheduler.utils.Site`, optional
        Name of the observory or a rubin_scheduler.utils.Site object.
        Default 'LSST'.
    """

    cols_added = ["PA"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        mjd_col="observationStartMJD",
        lst_col="observationStartLST",
        site="LSST",
    ):
        self.lst_col = lst_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        self.mjd_col = mjd_col
        self.site = Site(name=site)
        self.units = ["radians"]
        self.cols_req = [self.ra_col, self.dec_col, self.mjd_col, self.lst_col]
        self.ha_stacker = HourAngleStacker(lst_col=lst_col, ra_col=ra_col, degrees=self.degrees)

    def _run(self, sim_data, cols_present=False):
        # Equation from:
        # http://www.gb.nrao.edu/~rcreager/GBTMetrology/140ft/l0058/gbtmemo52/memo52.html
        # or
        # http://www.gb.nrao.edu/GBT/DA/gbtidl/release2pt9/contrib/contrib/parangle.pro
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data
        # Using the run method (not _run) means that if HA is present, it will
        # not be recalculated.
        sim_data = self.ha_stacker.run(sim_data)
        if self.degrees:
            dec = np.radians(sim_data[self.dec_col])
        else:
            dec = sim_data[self.dec_col]
        sim_data["PA"] = np.arctan2(
            np.sin(sim_data["HA"] * np.pi / 12.0),
            (
                np.cos(dec) * np.tan(self.site.latitude_rad)
                - np.sin(dec) * np.cos(sim_data["HA"] * np.pi / 12.0)
            ),
        )
        if self.degrees:
            sim_data["PA"] = np.degrees(sim_data["PA"])
        return sim_data


class OverheadStacker(BaseStacker):
    """Add time between visits in seconds.

    Parameters
    ----------
    max_gap : `float`, optional
        The maximum gap between observations, in minutes.
        Assume anything longer the dome has closed.
        Defaults to infinity.
    mjd_col : `str`, optional
        The name of the column with the observation start MJD.
        Defaults to "observationStartMJD".
    visit_time_col : `str`, optional
        The name of the column with the total visit time (on-sky plus
        shutter and other overheads).
        Defaults to "visitTime".
    exposure_time_col : `str`, optional
        The name of the column with the visit on-sky exposure time.
        Defaults to "visitExposureTime."
    """

    cols_added = ["overhead"]

    def __init__(
        self,
        max_gap=np.inf,
        mjd_col="observationStartMJD",
        visit_time_col="visitTime",
        exposure_time_col="visitExposureTime",
    ):
        # Set max_gap in minutes to match API of existing BruteOSFMetric
        self.max_gap = max_gap
        self.mjd_col = mjd_col
        self.visit_time_col = visit_time_col
        self.exposure_time_col = exposure_time_col
        self.units = ["seconds"]
        self.day_obs_mjd_stacker = DayObsMJDStacker(self.mjd_col)
        self.cols_req = [self.mjd_col, self.visit_time_col, self.exposure_time_col]
        self.cols_req.extend(self.day_obs_mjd_stacker.cols_req)
        self.cols_req = list(set(self.cols_req))

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data

        # Count overhead as any time from the shutter close of the previous
        # visit, to the shutter close of the current visit, minus the
        # current on-sky exposure time.
        # This is all non-exposure time required for this visit --
        # includes slew time to this target, any shutter overheads for this
        # visit, and any readouts not absorbed in slewtime for this visit.
        observation_end_mjd = sim_data[self.mjd_col] + sim_data[self.visit_time_col] / (24 * 60 * 60)
        overhead = (
            np.diff(observation_end_mjd, prepend=np.nan) * 24 * 60 * 60 - sim_data[self.exposure_time_col]
        )

        # Rough heuristic not to count downtime due to weather or instrument
        # problems as overhead.  A more reliable way to do this would be to
        # use other sources of weather data or problem reporting to identify
        # these gaps.
        # We might also explicitly want to look at the gaps that arise from
        # these problems in order to measure time lost to these causes.
        overhead[overhead > self.max_gap * 60] = np.nan

        # If the gap includes a change of night, it isn't really overhead. For
        # most reasonable finite values of self.max_gap, this is never
        # relevant, but it's sometimes useful to set max_gap to inifinity to
        # catch all delays in the night. (The comment above.) But, in
        # even in this case, we would still not want to include gaps between
        # nights.
        day_obs_mjd = self.day_obs_mjd_stacker.run(sim_data)["day_obs_mjd"]
        different_night = np.diff(day_obs_mjd, prepend=np.nan) != 0
        overhead[different_night] = np.nan

        sim_data[self.cols_added[0]] = overhead
        return sim_data



import os
import sys
import unittest
import warnings

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time
from rubin_scheduler.utils import Site, _alt_az_pa_from_ra_dec, calc_lmst

import rubin_sim.maf.stackers as stackers
from rubin_sim.data import get_data_dir
from rubin_sim.maf import get_sim_data

try:
    import lsst.summit.utils.efdUtils
except ModuleNotFoundError:
    pass

TEST_DB = "example_v3.4_0yrs.db"


class TestStackerClasses(unittest.TestCase):
    def setUp(self):
        # get some of the test data
        test_db = os.path.join(get_data_dir(), "tests", TEST_DB)
        query = "select * from observations limit 1000"
        self.test_data = get_sim_data(test_db, None, [], full_sql_query=query)

    def test_stackers_run(self):
        """Just run all of the stackers with our example data."""
        for stacker_class in stackers.BaseStacker.registry.values():
            stacker = stacker_class()
            stacker_name = stacker.__class__.__name__.lower()
            if stacker_name.startswith("sdss"):
                continue
            if isinstance(stacker, stackers.BaseMoStacker):
                continue
            try:
                stacker.run(self.test_data)
            except NotImplementedError:
                pass
            except:
                print(f"Failed at stacker {stacker.__class__.__name__}")
                raise

    def test_add_cols(self):
        """Test that we can add columns as expected."""
        data = np.zeros(90, dtype=list(zip(["alt"], [float])))
        data["alt"] = np.arange(0, 90)
        stacker = stackers.ZenithDistStacker(alt_col="alt", degrees=True)
        newcol = stacker.cols_added[0]
        # First - are the columns added if they are not there.
        data, cols_present = stacker._add_stacker_cols(data)
        self.assertEqual(cols_present, False)
        self.assertIn(newcol, data.dtype.names)
        # Next - if they are present, is that information passed back?
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            data, cols_present = stacker._add_stacker_cols(data)
            self.assertEqual(cols_present, True)

    def test_eq(self):
        """
        Test that stackers can be compared
        """
        s1 = stackers.ParallaxFactorStacker()
        s2 = stackers.ParallaxFactorStacker()
        assert s1 == s2

        # Test if they have numpy array atributes
        s1.ack = np.arange(10)
        s2.ack = np.arange(10)
        assert s1 == s2

        # Change the array and test
        s1.ack += 1
        assert s1 != s2

        s2 = stackers.NormAirmassStacker()
        assert s1 != s2

    def test_norm_airmass(self):
        """
        Test the normalized airmass stacker.
        """
        rng = np.random.RandomState(232)
        data = np.zeros(600, dtype=list(zip(["airmass", "fieldDec"], [float, float])))
        data["airmass"] = rng.random_sample(600)
        data["fieldDec"] = rng.random_sample(600) * np.pi - np.pi / 2.0
        data["fieldDec"] = np.degrees(data["fieldDec"])
        stacker = stackers.NormAirmassStacker(degrees=True)
        data = stacker.run(data)
        for i in np.arange(data.size):
            self.assertLessEqual(data["normairmass"][i], data["airmass"][i])
        self.assertLess(np.min(data["normairmass"] - data["airmass"]), 0)

    def test_parallax_factor(self):
        """
        Test the parallax factor.
        """

        data = np.zeros(
            600,
            dtype=list(
                zip(
                    ["fieldRA", "fieldDec", "observationStartMJD"],
                    [float, float, float],
                )
            ),
        )
        data["fieldRA"] = data["fieldRA"] + 0.1
        data["fieldDec"] = data["fieldDec"] - 0.1
        data["observationStartMJD"] = np.arange(data.size) + 49000.0
        stacker = stackers.ParallaxFactorStacker(degrees=True)
        data = stacker.run(data)
        self.assertLess(max(np.abs(data["ra_pi_amp"])), 1.1)
        self.assertLess(max(np.abs(data["dec_pi_amp"])), 1.1)
        self.assertLess(np.max(data["ra_pi_amp"] ** 2 + data["dec_pi_amp"] ** 2), 1.1)
        self.assertGreater(min(np.abs(data["ra_pi_amp"])), 0.0)
        self.assertGreater(min(np.abs(data["dec_pi_amp"])), 0.0)

    def _t_dither_range(self, diffsra, diffsdec, ra, dec, max_dither):
        self.assertLessEqual(np.abs(diffsra).max(), max_dither)
        self.assertLessEqual(np.abs(diffsdec).max(), max_dither)
        offsets = np.sqrt(diffsra**2 + diffsdec**2)
        self.assertLessEqual(offsets.max(), max_dither)
        self.assertGreater(diffsra.max(), 0)
        self.assertGreater(diffsdec.max(), 0)
        self.assertLess(diffsra.min(), 0)
        self.assertLess(diffsdec.min(), 0)

    def _t_dither_per_night(self, diffsra, diffsdec, ra, dec, nights):
        n = np.unique(nights)
        for ni in n:
            match = np.where(nights == ni)[0]
            dra_on_night = np.abs(np.diff(diffsra[match]))
            ddec_on_night = np.abs(np.diff(diffsdec[match]))
            if dra_on_night.max() > 0.0005:
                print(ni)
                m = np.where(dra_on_night > 0.0005)[0]
                print(diffsra[match][m])
                print(ra[match][m])
                print(dec[match][m])
                print(dra_on_night[m])
            self.assertAlmostEqual(dra_on_night.max(), 0)
            self.assertAlmostEqual(ddec_on_night.max(), 0)

    def test_ha_stacker(self):
        """Test the Hour Angle stacker"""
        data = np.zeros(100, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data["observationStartLST"] = np.arange(100) / 99.0 * np.pi * 2
        stacker = stackers.HourAngleStacker(degrees=True)
        data = stacker.run(data)
        # Check that data is always wrapped
        self.assertLess(np.max(data["HA"]), 12.0)
        self.assertGreater(np.min(data["HA"]), -12.0)
        # Check that HA is zero if lst == RA
        data = np.zeros(1, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data = stacker.run(data)
        self.assertEqual(data["HA"], 0.0)
        data = np.zeros(1, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data["observationStartLST"] = 20.0
        data["fieldRA"] = 20.0
        data = stacker.run(data)
        self.assertEqual(data["HA"], 0.0)
        # Check a value
        data = np.zeros(1, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data["observationStartLST"] = 0.0
        data["fieldRA"] = np.degrees(np.pi / 2.0)
        data = stacker.run(data)
        np.testing.assert_almost_equal(data["HA"], -6.0)

    def test_pa_stacker(self):
        """Test the parallacticAngleStacker"""
        data = np.zeros(
            100,
            dtype=list(
                zip(
                    [
                        "observationStartMJD",
                        "fieldDec",
                        "fieldRA",
                        "observationStartLST",
                    ],
                    [float] * 4,
                )
            ),
        )
        data["observationStartMJD"] = np.arange(100) * 0.2 + 50000
        site = Site(name="LSST")
        data["observationStartLST"] = calc_lmst(data["observationStartMJD"], site.longitude_rad)
        data["observationStartLST"] = data["observationStartLST"] * 180.0 / 12.0
        stacker = stackers.ParallacticAngleStacker(degrees=True)
        data = stacker.run(data)
        # Check values are in good range
        assert data["PA"].max() <= 180
        assert data["PA"].min() >= -180

        # Check compared to the util
        check_pa = []
        ras = np.radians(data["fieldRA"])
        decs = np.radians(data["fieldDec"])
        for ra, dec, mjd in zip(ras, decs, data["observationStartMJD"]):
            alt, az, pa = _alt_az_pa_from_ra_dec(ra, dec, mjd, site.longitude_rad, site.latitude_rad)

            check_pa.append(pa)
        check_pa = np.degrees(check_pa)

        np.testing.assert_array_almost_equal(data["PA"], check_pa, decimal=0)

    def test_galactic_stacker(self):
        """
        Test the galactic coordinate stacker
        """
        ra, dec = np.degrees(
            np.meshgrid(np.arange(0, 2.0 * np.pi, 0.1), np.arange(-np.pi / 2, np.pi / 2, 0.1))
        )
        ra = np.ravel(ra)
        dec = np.ravel(dec)
        data = np.zeros(ra.size, dtype=list(zip(["ra", "dec"], [float] * 2)))
        data["ra"] += ra
        data["dec"] += dec
        s = stackers.GalacticStacker(ra_col="ra", dec_col="dec")
        new_data = s.run(data)
        c = SkyCoord(ra=ra * u.deg, dec=dec * u.deg).transform_to("galactic")
        expected_l, expected_b = c.l.rad, c.b.rad
        np.testing.assert_array_equal(new_data["gall"], expected_l)
        np.testing.assert_array_equal(new_data["galb"], expected_b)

        # Check that we have all the quadrants populated
        q1 = np.where((new_data["gall"] < np.pi) & (new_data["galb"] < 0.0))[0]
        q2 = np.where((new_data["gall"] < np.pi) & (new_data["galb"] > 0.0))[0]
        q3 = np.where((new_data["gall"] > np.pi) & (new_data["galb"] < 0.0))[0]
        q4 = np.where((new_data["gall"] > np.pi) & (new_data["galb"] > 0.0))[0]
        assert q1.size > 0
        assert q2.size > 0
        assert q3.size > 0
        assert q4.size > 0

    def test_ecliptic_stacker(self):
        ra, dec = np.degrees(
            np.meshgrid(np.arange(0, 2.0 * np.pi, 0.1), np.arange(-np.pi / 2, np.pi / 2, 0.1))
        )
        ra = np.ravel(ra)
        dec = np.ravel(dec)
        data = np.zeros(ra.size, dtype=list(zip(["ra", "dec"], [float] * 2)))
        data["ra"] += ra
        data["dec"] += dec
        s = stackers.EclipticStacker(ra_col="ra", dec_col="dec", degrees=True)
        _ = s.run(data)

        data = np.zeros(ra.size, dtype=list(zip(["ra", "dec"], [float] * 2)))
        data["ra"] += ra
        data["dec"] += dec
        s = stackers.EclipticStacker(ra_col="ra", dec_col="dec", degrees=True, subtract_sun_lon=False)
        _ = s.run(data)

    def test_teff_stacker(self):
        rng = np.random.default_rng(seed=6563)
        num_points = 5
        data = np.zeros(
            num_points,
            dtype=list(zip(["fiveSigmaDepth", "filter", "visitExposureTime"], [float, (np.str_, 1), float])),
        )
        data["fiveSigmaDepth"] = 23 + rng.random(num_points)
        data["filter"] = ["g"] * num_points
        data["visitExposureTime"] = [30] * num_points

        stacker = stackers.TeffStacker("fiveSigmaDepth", "filter", "visitExposureTime")
        value = stacker.run(data)
        np.testing.assert_array_equal(value["fiveSigmaDepth"], value["fiveSigmaDepth"])
        assert np.all(0.1 < value["t_eff"]) and np.all(value["t_eff"] < 10)

    def test_observation_start_datetime64_stacker(self):
        rng = np.random.default_rng(seed=6563)
        num_points = 5
        data = np.zeros(
            num_points,
            dtype=list(zip(["observationStartMJD"], [float])),
        )
        data["observationStartMJD"] = 61000 + 3000 * rng.random(num_points)

        stacker = stackers.ObservationStartDatetime64Stacker("observationStartMJD")
        value = stacker.run(data)
        recovered_mjd = Time(value["observationStartDatetime64"], format="datetime64").mjd
        assert np.allclose(recovered_mjd, data["observationStartMJD"])

    def test_day_obs_stackers(self):
        rng = np.random.default_rng(seed=6563)
        num_points = 5
        obs_start_mjds = 61000 + 3000 * rng.random(num_points)

        data = np.zeros(
            num_points,
            dtype=list(zip(["observationStartMJD"], [float])),
        )
        data["observationStartMJD"] = obs_start_mjds

        try:
            summit_get_day_obs = lsst.summit.utils.efdUtils.getDayObsForTime
        except NameError:
            summit_get_day_obs = None

        day_obs_int_stacker = stackers.DayObsStacker("observationStartMJD")
        day_obs_int = day_obs_int_stacker.run(data)
        assert np.all(day_obs_int["dayObs"] > 20200000)
        assert np.all(day_obs_int["dayObs"] < 20500000)
        if summit_get_day_obs is not None:
            summit_day_obs_int = np.array([summit_get_day_obs(Time(m, format="mjd")) for m in obs_start_mjds])
            np.testing.assert_array_equal(day_obs_int["dayObs"], summit_day_obs_int)

        day_obs_mjd_stacker = stackers.DayObsMJDStacker("observationStartMJD")
        new_data = day_obs_mjd_stacker.run(data)
        # If there were no offset, all values would be between 0 and 1.
        # With the 0.5 day offset specified in SITCOMTN-32, this becomes
        # 0.5 and 1.5.
        assert np.all((obs_start_mjds - new_data["day_obs_mjd"]) <= 1.5)
        assert np.all((obs_start_mjds - new_data["day_obs_mjd"]) >= 0.5)

        day_obs_iso_stacker = stackers.DayObsISOStacker("observationStartMJD")
        _ = day_obs_iso_stacker.run(data)

    def test_overhead_stacker(self):
        num_visits = 10
        num_first_night_visits = 5

        start_mjd = 61000.2
        exptime = 30.0

        exposure_times = np.full(num_visits, exptime, dtype=float)

        rng = np.random.default_rng(seed=6563)
        overheads = 2.0 + rng.random(num_visits)

        visit_times = overheads + exposure_times
        observation_end_mjds = start_mjd + np.cumsum(visit_times) / (24 * 60 * 60)
        mjds = observation_end_mjds - visit_times / (24 * 60 * 60)
        mjds[num_first_night_visits:] += 1

        data = np.core.records.fromarrays(
            (mjds, exposure_times, visit_times),
            dtype=np.dtype(
                [
                    ("observationStartMJD", mjds.dtype),
                    ("visitExposureTime", exposure_times.dtype),
                    ("visitTime", visit_times.dtype),
                ]
            ),
        )

        overhead_stacker = stackers.OverheadStacker()
        new_data = overhead_stacker.run(data)
        measured_overhead = new_data["overhead"]

        # There is no visit before the first, so the first overhead should be
        # nan.
        self.assertTrue(np.isnan(measured_overhead[0]))

        # Test that the gap between nights is nan.
        self.assertTrue(np.isnan(measured_overhead[num_first_night_visits]))

        # Make sure there are no unexpected nans
        self.assertEqual(np.count_nonzero(np.isnan(measured_overhead)), 2)

        # Make sure all measured values are correct
        these_gaps = ~np.isnan(measured_overhead)
        self.assertTrue(np.allclose(measured_overhead[these_gaps], overheads[these_gaps]))

    def testHelp(self):
        # essentially, test that every stacker has cols_req and cols_added
        with open(os.devnull, "w") as new_target:
            sys.stdout = new_target
            stackers.BaseStacker.help(doc=True)


if __name__ == "__main__":
    unittest.main()



__all__ = ("OptimalM5Metric",)

import warnings

import numpy as np

from .base_metric import BaseMetric
from .simple_metrics import Coaddm5Metric


class OptimalM5Metric(BaseMetric):
    """Compare the co-added depth of the survey to one where
    all the observations were taken on the meridian.

    Parameters
    ----------
    m5_col : str ('fiveSigmaDepth')
        Column name that contains the five-sigma limiting depth of
        each observation
    opt_m5_col : str ('m5Optimal')
        The column name of the five-sigma-limiting depth if the
        observation had been taken on the meridian.
    normalize : bool (False)
        If False, metric returns how many more observations would need
        to be taken to reach the optimal depth.  If True, the number
        is normalized by the total number of observations already taken
        at that position.
    mag_diff : bool (False)
        If True, metric returns the magnitude difference between the
        achieved coadded depth and the optimal coadded depth.

    Returns
    --------
    numpy.array

    If mag_diff is True, returns the magnitude difference between the
    optimal and actual coadded depth.  If normalize is False
    (default), the result is the number of additional observations
    (taken at the median depth) the survey needs to catch up to
    optimal.  If normalize is True, the result is divided by the
    number of observations already taken. So if a 10-year survey
    returns 20%, it would need to run for 12 years to reach the same
    depth as a 10-year meridian survey.

    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        opt_m5_col="m5Optimal",
        filter_col="filter",
        mag_diff=False,
        normalize=False,
        **kwargs,
    ):
        if normalize:
            self.units = "% behind"
        else:
            self.units = "N visits behind"
        if mag_diff:
            self.units = "mags"
        super(OptimalM5Metric, self).__init__(
            col=[m5_col, opt_m5_col, filter_col], units=self.units, **kwargs
        )
        self.m5_col = m5_col
        self.opt_m5_col = opt_m5_col
        self.normalize = normalize
        self.filter_col = filter_col
        self.mag_diff = mag_diff
        self.coadd_regular = Coaddm5Metric(m5_col=m5_col)
        self.coadd_optimal = Coaddm5Metric(m5_col=opt_m5_col)

    def run(self, data_slice, slice_point=None):
        filters = np.unique(data_slice[self.filter_col])
        if np.size(filters) > 1:
            warnings.warn(
                "OptimalM5Metric does not make sense mixing filters. Currently using filters " + str(filters)
            )
        regular_depth = self.coadd_regular.run(data_slice)
        optimal_depth = self.coadd_optimal.run(data_slice)
        if self.mag_diff:
            return optimal_depth - regular_depth

        median_single = np.median(data_slice[self.m5_col])

        # Number of additional median observations to get as deep as optimal
        result = (10.0 ** (0.8 * optimal_depth) - 10.0 ** (0.8 * regular_depth)) / (
            10.0 ** (0.8 * median_single)
        )

        if self.normalize:
            result = result / np.size(data_slice) * 100.0

        return result
